"""ASCII representations of retained strata."""

from ._col_to_ascii import col_to_ascii

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "col_to_ascii",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import pytest

from hstrat.hstrat import recency_proportional_resolution_curbed_algo


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(size_curb):
    spec = recency_proportional_resolution_curbed_algo.PolicySpec(size_curb)
    assert spec == spec
    assert spec == recency_proportional_resolution_curbed_algo.PolicySpec(
        size_curb
    )
    assert not spec == recency_proportional_resolution_curbed_algo.PolicySpec(
        size_curb + 1
    )


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_init(size_curb):
    spec = recency_proportional_resolution_curbed_algo.PolicySpec(size_curb)
    assert spec._size_curb == size_curb


def test_GetAlgoIdentifier():
    size_curb = 42
    spec = recency_proportional_resolution_curbed_algo.PolicySpec(size_curb)
    assert spec.GetAlgoIdentifier()


def test_GetAlgoTitle():
    size_curb = 42
    spec = recency_proportional_resolution_curbed_algo.PolicySpec(size_curb)
    assert spec.GetAlgoTitle()


def test_repr():
    size_curb = 42
    spec = recency_proportional_resolution_curbed_algo.PolicySpec(size_curb)
    assert str(size_curb) in repr(spec)
    assert spec.GetAlgoIdentifier() in repr(spec)


def test_str():
    size_curb = 42
    spec = recency_proportional_resolution_curbed_algo.PolicySpec(size_curb)
    assert str(size_curb) in str(spec)
    assert spec.GetAlgoTitle() in str(spec)



import pytest
from slugify import slugify
from teeplot import teeplot as tp

from hstrat import hstrat
from hstrat._auxiliary_lib import release_cur_mpl_fig


@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(10),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test(policy):
    hstrat.mrca_uncertainty_relative_barplot(policy, 100, do_show=False)
    release_cur_mpl_fig()
    hstrat.mrca_uncertainty_relative_barplot(policy, 10, do_show=False)
    release_cur_mpl_fig()


@pytest.mark.heavy
def test_docplots(docplot_policy):
    tp.tee(
        hstrat.mrca_uncertainty_relative_barplot,
        docplot_policy,
        256,
        teeplot_outattrs={
            "policy": slugify(str(docplot_policy)),
            "num_generations": "256",
        },
        teeplot_transparent=False,
    )
    release_cur_mpl_fig()



import anytree

from ._AnyTreeFastPreOrderIter import AnyTreeFastPreOrderIter


def anytree_cardinality(node: anytree.Node) -> int:
    """Count the number of nodes in an anytree tree rooted at the given node.

    Parameters
    ----------
    node : anytree.Node
        The root node of the Anytree tree to compute the size of.

    Returns
    -------
    int
        The number of nodes in the tree rooted at the given node.
    """
    return sum(1 for _ in AnyTreeFastPreOrderIter(node))



import typing

from .._auxiliary_lib import get_hstrat_version, log_once_in_a_row
from ._impl import policy_from_record
from ._policy_to_records import policy_to_records


def policy_from_records(records: typing.Dict) -> typing.Callable:
    """Deserialize a stratum retention policy from a dict composed of builtin
    data types.
    """
    if get_hstrat_version() != records["hstrat_version"]:
        log_once_in_a_row(
            f"""policy_from_records version mismatch, record is version {
                records['hstrat_version']
            } and software is version {
                get_hstrat_version()
            }"""
        )
        # return policy without any further safety checks
        return policy_from_record(records["policy"])

    deserialized_policy = policy_from_record(records["policy"])

    # safety check metadata
    deserialized_policy_records = policy_to_records(deserialized_policy)
    if not all(
        k not in records or v == records[k]
        for k, v in deserialized_policy_records.items()
    ):
        log_once_in_a_row(
            f"""policy_from_records metadata mismatch, records are {
                records
            } and records from deserialized policy are {
                deserialized_policy_records
            }"""
        )

    return deserialized_policy



import anytree

from ...._auxiliary_lib import render_to_base64url


class TrieLeafNode(anytree.NodeMixin):
    """A leaf node in a  phylogenetic reconstruction trie.

    Represents an single, specific extant annotation among the population
    subject to phylogenetic inference.

    Parameters
    ----------
    parent : TrieInnerNode
        The parent node of this leaf node.
    taxon_label : str
        The taxon label for this leaf node.

    Attributes
    ----------
    taxon_label : str
        The taxon label for this leaf node.
    taxon : str
        Synonym for taxon_label.
    """

    taxon_label: str

    def __init__(
        self: "TrieLeafNode",
        parent: "TrieInnerNode",
        taxon_label: str,
    ) -> None:
        """Initialize a new TrieLeafNode.

        Parameters
        ----------
        parent : TrieInnerNode
            The parent node of this leaf node.
        taxon_label : str
            The taxon label for this leaf node.
        """
        self.parent = parent
        self.taxon_label = taxon_label

    @property
    def taxon(self: "TrieLeafNode") -> str:
        """Return the taxon label for this leaf node."""
        return self.taxon_label

    @property
    def rank(self: "TrieLeafNode") -> int:
        """Return the origin time for this leaf node."""
        return self.parent.rank

    def __repr__(self: "TrieLeafNode") -> str:
        """Return a string representation of this leaf node."""
        return f"""{self.taxon_label} @ {
            render_to_base64url(id(self) % 8192)
        }"""



"""Implementation helpers."""

from ._TrieInnerNode import TrieInnerNode
from ._TrieLeafNode import TrieLeafNode
from ._append_genesis_organism import append_genesis_organism
from ._build_tree_biopython_distance import build_tree_biopython_distance
from ._build_trie_from_artifacts import build_trie_from_artifacts
from ._estimate_origin_times import estimate_origin_times
from ._find_chronological_root import find_chronological_root
from ._find_chronological_roots import find_chronological_roots
from ._time_calibrate_tree import time_calibrate_tree

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "TrieInnerNode",
    "TrieLeafNode",
    "append_genesis_organism",
    "build_tree_biopython_distance",
    "build_trie_from_artifacts",
    "estimate_origin_times",
    "find_chronological_root",
    "find_chronological_roots",
    "time_calibrate_tree",
]



import numpy as np
import pytest

from hstrat.hstrat import recency_proportional_resolution_curbed_algo


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_policy_consistency(size_curb, time_sequence):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.CalcNumStrataRetainedUpperBound(
        spec
    )
    for num_strata_deposited in time_sequence:
        policy_requirement = policy.CalcNumStrataRetainedExact(
            num_strata_deposited,
        )
        for which in (
            instance,
            recency_proportional_resolution_curbed_algo.CalcNumStrataRetainedUpperBound(
                spec
            ),
        ):
            assert (
                which(
                    policy,
                    num_strata_deposited,
                )
                >= policy_requirement
            )


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(size_curb):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.CalcNumStrataRetainedUpperBound(
        spec
    )

    assert instance == instance
    assert (
        instance
        == recency_proportional_resolution_curbed_algo.CalcNumStrataRetainedUpperBound(
            spec,
        )
    )
    assert instance is not None



from unittest.mock import MagicMock
import warnings

try:
    with warnings.catch_warnings(record=True) as w:
        import Bio.Phylo.TreeConstruction as BioPhyloTree

        _ = BioPhyloTree

        if w:  # pragma: no cover
            warnings.warn(
                "ImportWarning: warnings from Biopython import were silenced.",
                ImportWarning,
            )
except:  # pragma: no cover
    warnings.warn(
        "ImportWarning: Bio.Phylo.TreeConstruction import failed; "
        "inserting a no-op mock for BioPhyloTree."
    )
    BioPhyloTree = MagicMock()



import copy

import anytree


def anynode_deepcopy_except_neighbors(
    node: anytree.AnyNode,
) -> anytree.AnyNode:
    """Create a deep copy of an AnyNode object, excluding its parents and
    children.

    Parameters
    ----------
    node : anytree.AnyNode
        The AnyNode object to be copied.

    Returns
    -------
    anytree.AnyNode
        A new AnyNode object with the same attributes and values as the
        original, but with no parent or children.
    """
    hasattr_parent = hasattr(node, "_NodeMixin__parent")

    # temporarily detach parent and children
    detached_children, node._NodeMixin__children = (
        node._NodeMixin__children,
        list(),
    )
    if hasattr_parent:
        detached_parent, node._NodeMixin__parent = (
            node._NodeMixin__parent,
            None,
        )

    res = copy.deepcopy(node)

    # reattach parent and children
    node._NodeMixin__children = detached_children
    if hasattr_parent:
        node._NodeMixin__parent = detached_parent

    return res



import typing

from ..._detail import PolicyCouplerBase
from ..._impl import CalcMrcaUncertaintyAbsUpperBoundWorstCase
from .._PolicySpec import PolicySpec


class CalcMrcaUncertaintyAbsUpperBound:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsUpperBound",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyAbsUpperBound",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsUpperBound",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> int:
        """At most, how much absolute uncertainty to estimate rank of MRCA?
        Inclusive."""
        spec = policy.GetSpec()

        res = spec.GetFixedResolution() - 1

        return min(
            res,
            CalcMrcaUncertaintyAbsUpperBoundWorstCase()(
                policy,
                first_num_strata_deposited,
                second_num_strata_deposited,
                actual_rank_of_mrca,
            ),
        )



import numpy as np


def min_array_dtype(arr: np.array) -> np.dtype:
    """Determine the most compact numpy dtype required to represent the values
    in the input array.

    Parameters
    ----------
    arr : numpy.ndarray
        Input array.

    Returns
    -------
    numpy.dtype
        The most compact dtype required to represent the values in the input
        array.

    Notes
    -----
    If the input array is empty, the function returns the result of
    `np.min_scalar_type(0)` as the default dtype.

    Like `np.min_scalar_type` this method ignores loss of precision at values
    close to 0 (i.e., subnormal numbers).

    Examples
    --------
    >>> arr = np.array([-129, 128])
    >>> dtype = min_array_dtype(arr)
    >>> print(dtype)
    int8
    """
    if len(arr) == 0:
        return np.min_scalar_type(0)

    min_, max_ = arr.min(), arr.max()
    if min_ >= 0:
        return np.min_scalar_type(max_)

    dtype_min = np.min_scalar_type(min_)
    conv = dtype_min.type(max_)
    if conv == max_ and np.isfinite(conv):
        return dtype_min
    else:
        assert abs(min_) < max_
        return np.min_scalar_type(-max_)



import more_itertools as mit
import pytest
import typing_extensions

from hstrat import hstrat


@pytest.mark.parametrize(
    "differentia_bit_width",
    [
        1,
        8,
        16,
        32,
        64,
    ],
)
@pytest.mark.parametrize(
    "num_strata",
    [
        0,
        1,
        2,
        7,
        8,
        15,
        27,
        32,
    ],
)
def test_pack_differentiae_bytes(differentia_bit_width, num_strata):
    original_strata = [
        hstrat.HereditaryStratum(
            differentia_bit_width=differentia_bit_width,
            deposition_rank=i,
        )
        for i in range(num_strata)
    ]
    packed_differentiae = hstrat.pack_differentiae_bytes(
        original_strata,
        differentia_bit_width,
    )
    assert isinstance(packed_differentiae, typing_extensions.Buffer)

    unpacked_differentiae = hstrat.unpack_differentiae_bytes(
        packed_differentiae,
        differentia_bit_width,
    )

    for stratum, differentia in mit.zip_equal(
        original_strata,
        unpacked_differentiae,
    ):
        assert stratum.GetDifferentia() == differentia


@pytest.mark.parametrize(
    "differentia_bit_width",
    [
        1,
        8,
        16,
        32,
        64,
    ],
)
@pytest.mark.parametrize(
    "num_strata",
    [
        0,
        1,
        2,
        7,
        8,
        15,
        27,
        32,
    ],
)
def test_pack_differentiae_bytes_specify_num_differentia_packed(
    differentia_bit_width, num_strata
):
    original_strata = [
        hstrat.HereditaryStratum(
            differentia_bit_width=differentia_bit_width,
            deposition_rank=i,
        )
        for i in range(num_strata)
    ]
    packed_differentiae = hstrat.pack_differentiae_bytes(
        original_strata,
        differentia_bit_width,
        always_omit_num_padding_bits_header=True,
    )
    assert isinstance(packed_differentiae, typing_extensions.Buffer)

    unpacked_differentiae = hstrat.unpack_differentiae_bytes(
        packed_differentiae,
        differentia_bit_width,
        num_packed_differentia=len(original_strata),
    )

    for stratum, differentia in mit.zip_equal(
        original_strata,
        unpacked_differentiae,
    ):
        assert stratum.GetDifferentia() == differentia



import typing

from .._auxiliary_lib import get_hstrat_version
from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._pack_differentiae_str import pack_differentiae_str
from ._policy_to_records import policy_to_records


def col_to_records(column: HereditaryStratigraphicColumn) -> typing.Dict:
    """Serialize a `HereditaryStratigraphicColumn` to a dict composed of
    builtin types."""
    differentia_bit_width = column.GetStratumDifferentiaBitWidth()
    packed_differentiae = pack_differentiae_str(
        column.IterRetainedStrata(),
        differentia_bit_width,
    )

    policy_records = policy_to_records(column._stratum_retention_policy)
    res = {
        **policy_records,
        **{
            "num_strata_deposited": column.GetNumStrataDeposited(),
            "differentiae": packed_differentiae,
            "differentia_bit_width": differentia_bit_width,
            "hstrat_version": get_hstrat_version(),
        },
    }
    if not column._ShouldOmitStratumDepositionRank():
        res["stratum_deposition_ranks"] = [
            stratum.GetDepositionRank()
            for stratum in column.IterRetainedStrata()
        ]
        assert (
            len(res["stratum_deposition_ranks"])
            <= column.GetNumStrataDeposited()
        )

    if column.HasAnyAnnotations():
        res["stratum_annotations"] = [
            stratum.GetAnnotation() for stratum in column.IterRetainedStrata()
        ]

    return res



import collections
import itertools as it

import alifedata_phyloinformatics_convert as apc
import numpy as np
import pandas as pd
import pytest

from hstrat import hstrat


def _compare_compiled_phylogenies(
    decantless_phylogeny_df: pd.DataFrame,
    decanting_phylogeny_df: pd.DataFrame,
    share_common_ancestor: bool,
) -> None:
    # compile tree tracked without decanting
    decantless_trees = apc.alife_dataframe_to_dendropy_trees(
        decantless_phylogeny_df,
    )
    if share_common_ancestor:
        assert len(decantless_trees) == 1

    # compile tree tracked with decanting
    assert len(decanting_phylogeny_df) == len(
        decanting_phylogeny_df["id"].unique()
    )

    decanting_trees = apc.alife_dataframe_to_dendropy_trees(
        decanting_phylogeny_df,
    )
    assert sum(
        1
        for decanting_tree in decanting_trees
        for __ in decanting_tree.leaf_node_iter()
    ) == sum(
        1
        for decantless_tree in decantless_trees
        for __ in decantless_tree.leaf_node_iter()
    )
    assert (
        len(
            set(
                node.level()
                for decanting_tree in decantless_trees
                for node in decanting_tree.leaf_node_iter()
            )
        )
        == 1
    )

    assert len(decanting_trees) == len(decantless_trees)

    # setup taxa in both trees
    for trees in decanting_trees, decantless_trees:
        for i, tree in enumerate(trees):
            for j, n in enumerate(tree):
                n.taxon = tree.taxon_namespace.new_taxon(f"{i} {j}")

    # check phylogenetic trees with and without decanting are identical
    assert collections.Counter(
        node.level()
        for decantless_tree in decantless_trees
        for node in decantless_tree
    ) == collections.Counter(
        node.level()
        for decanting_tree in decanting_trees
        for node in decanting_tree
    )
    assert collections.Counter(
        decantless_tree.mrca(taxa=[n1.taxon, n2.taxon]).level()
        for decantless_tree in decantless_trees
        for n1, n2 in it.combinations(decantless_tree.leaf_node_iter(), 2)
    ) == collections.Counter(
        decanting_tree.mrca(taxa=[n1.taxon, n2.taxon]).level()
        for decanting_tree in decanting_trees
        for n1, n2 in it.combinations(decanting_tree.leaf_node_iter(), 2)
    )


@pytest.mark.parametrize(
    "population_size",
    [
        20,
        100,
    ],
)
@pytest.mark.parametrize(
    "share_common_ancestor",
    [
        True,
        False,
    ],
)
def test_DecantingPhyloTracker(population_size, share_common_ancestor):

    # setup population and perfect trackers
    common_ancestor = hstrat.PerfectBacktrackHandle()
    handle_population = (
        [common_ancestor.CreateDescendant() for __ in range(population_size)]
        if share_common_ancestor
        else [hstrat.PerfectBacktrackHandle() for __ in range(population_size)]
    )
    tracker = hstrat.DecantingPhyloTracker(
        population_size=population_size,
        share_common_ancestor=share_common_ancestor,
    )

    # evolve fixed-size population with random selection
    for epoch_length in 0, 3, 50, 15, 5:
        for generation in range(epoch_length):
            parent_locs = np.random.randint(
                population_size, size=population_size
            )
            tracker.ElapseGeneration(parent_locs)
            handle_population = [
                handle_population[loc].CreateDescendant()
                for loc in parent_locs
            ]

        _compare_compiled_phylogenies(
            hstrat.compile_perfect_backtrack_phylogeny(handle_population),
            tracker.CompilePhylogeny(),
            share_common_ancestor,
        )



"""Functions to reconstruct a phylogenetic tree from extant hereditary strata."""

from . import trie_postprocess
from ._build_tree import build_tree
from ._build_tree_nj import build_tree_nj
from ._build_tree_trie import build_tree_trie
from ._build_tree_trie_ensemble import build_tree_trie_ensemble
from ._build_tree_upgma import build_tree_upgma
from .trie_postprocess import *  # noqa: F401

__all__ = [
    "build_tree_nj",
    "build_tree_trie",
    "build_tree_trie_ensemble",
    "build_tree_upgma",
    "build_tree",
    "trie_postprocess",
] + trie_postprocess.__all__

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder(
    [
        build_tree_nj,
        build_tree_trie,
        build_tree_trie_ensemble,
        build_tree_upgma,
        build_tree,
        trie_postprocess,
    ],
    __name__,
)
del _launder  # prevent name from leaking



import itertools as it
import random

import pytest

from hstrat import hstrat
from hstrat.juxtaposition._impl_column import (
    iter_ranks_of_retained_commonality_between,
)


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_asynchronous(
    retention_policy,
    ordered_store,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert [
                *iter_ranks_of_retained_commonality_between(first, second)
            ] == [*iter_ranks_of_retained_commonality_between(second, first)]

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy_2b,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    ordered_store,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert [
                *iter_ranks_of_retained_commonality_between(first, second)
            ] == [*iter_ranks_of_retained_commonality_between(second, first)]

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


# @pytest.mark.parametrize(
#     "retention_policy1",
#     [
#         hstrat.perfect_resolution_algo.Policy(),
#         hstrat.nominal_resolution_algo.Policy(),
#         hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
#     ],
# )
# @pytest.mark.parametrize(
#     "retention_policy2",
#     [
#         hstrat.perfect_resolution_algo.Policy(),
#         hstrat.nominal_resolution_algo.Policy(),
#         hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
#     ],
# )
# @pytest.mark.parametrize(
#     "ordered_store",
#     [
#         hstrat.HereditaryStratumOrderedStoreDict,
#         hstrat.HereditaryStratumOrderedStoreList,
#         hstrat.HereditaryStratumOrderedStoreTree,
#     ],
# )
# def test_scenario_no_mrca(
#     retention_policy1,
#     retention_policy2,
#     ordered_store,
# ):
#     first = hstrat.HereditaryStratigraphicColumn(
#         stratum_ordered_store=ordered_store,
#         stratum_retention_policy=retention_policy1,
#     )
#     second = hstrat.HereditaryStratigraphicColumn(
#         stratum_ordered_store=ordered_store,
#         stratum_retention_policy=retention_policy2,
#     )
#
#     for generation in range(100):
#         assert (
#             hstrat.calc_rank_of_last_retained_commonality_between(
#                 first, second
#             )
#             is None
#         )
#         assert (
#             hstrat.calc_rank_of_last_retained_commonality_between(
#                 second, first
#             )
#             is None
#         )
#
#         first.DepositStratum()
#         second.DepositStratum()
#

# @pytest.mark.parametrize(
#     "retention_policy",
#     [
#         hstrat.perfect_resolution_algo.Policy(),
#         hstrat.nominal_resolution_algo.Policy(),
#         hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
#     ],
# )
# @pytest.mark.parametrize(
#     "ordered_store",
#     [
#         hstrat.HereditaryStratumOrderedStoreDict,
#         hstrat.HereditaryStratumOrderedStoreList,
#         hstrat.HereditaryStratumOrderedStoreTree,
#     ],
# )
# def test_scenario_no_divergence(retention_policy, ordered_store):
#     column = hstrat.HereditaryStratigraphicColumn(
#         stratum_ordered_store=ordered_store,
#         stratum_retention_policy=retention_policy,
#     )
#
#     for generation in range(100):
#         assert (
#             hstrat.calc_rank_of_last_retained_commonality_between(
#                 column, column
#             )
#             == generation
#         )
#
#         column.DepositStratum()
#
#
# @pytest.mark.parametrize(
#     "retention_policy",
#     [
#         hstrat.perfect_resolution_algo.Policy(),
#         hstrat.nominal_resolution_algo.Policy(),
#         hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
#     ],
# )
# @pytest.mark.parametrize(
#     "ordered_store",
#     [
#         hstrat.HereditaryStratumOrderedStoreDict,
#         hstrat.HereditaryStratumOrderedStoreList,
#         hstrat.HereditaryStratumOrderedStoreTree,
#     ],
# )
# def test_scenario_partial_even_divergence(retention_policy, ordered_store):
#     first = hstrat.HereditaryStratigraphicColumn(
#         stratum_ordered_store=ordered_store,
#         stratum_retention_policy=retention_policy,
#     )
#
#     for generation in range(100):
#         first.DepositStratum()
#
#     second = first.Clone()
#
#     first.DepositStratum()
#     second.DepositStratum()
#
#     for generation in range(101, 200):
#         assert (
#             0
#             <= hstrat.calc_rank_of_last_retained_commonality_between(
#                 first, second
#             )
#             <= 100
#         )
#
#         first.DepositStratum()
#         second.DepositStratum()
#
#
# @pytest.mark.parametrize(
#     "retention_policy",
#     [
#         hstrat.perfect_resolution_algo.Policy(),
#         hstrat.nominal_resolution_algo.Policy(),
#         hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
#     ],
# )
# @pytest.mark.parametrize(
#     "ordered_store",
#     [
#         hstrat.HereditaryStratumOrderedStoreDict,
#         hstrat.HereditaryStratumOrderedStoreList,
#         pytest.param(
#             hstrat.HereditaryStratumOrderedStoreTree,
#             marks=pytest.mark.heavy,
#         ),
#     ],
# )
# def test_scenario_partial_uneven_divergence(retention_policy, ordered_store):
#     first = hstrat.HereditaryStratigraphicColumn(
#         stratum_ordered_store=ordered_store,
#         stratum_retention_policy=retention_policy,
#     )
#
#     for generation in range(100):
#         first.DepositStratum()
#
#     second = first.Clone()
#
#     first.DepositStratum()
#
#     for generation in range(101, 200):
#         assert (
#             0
#             <= hstrat.calc_rank_of_last_retained_commonality_between(
#                 first, second
#             )
#             <= 100
#         )
#
#         first.DepositStratum()
#
#     second.DepositStratum()
#
#     for generation in range(101, 200):
#         assert (
#             0
#             <= hstrat.calc_rank_of_last_retained_commonality_between(
#                 first, second
#             )
#             <= 100
#         )
#
#         second.DepositStratum()
#
#
# @pytest.mark.parametrize(
#     "retention_policy",
#     [
#         hstrat.perfect_resolution_algo.Policy(),
#         hstrat.nominal_resolution_algo.Policy(),
#         hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
#     ],
# )
# @pytest.mark.parametrize(
#     "differentia_width",
#     [1, 2, 8],
# )
# def _do_test_IterRanksOfRetainedCommonalityBetweenGenerich1(
#     ordered_store,
#     differentia_width,
# ):
#     column = hstrat.HereditaryStratigraphicColumn(
#         stratum_differentia_bit_width=differentia_width,
#         stratum_ordered_store=ordered_store,
#     )
#
#     for generation in range(100):
#         column.DepositStratum()
#
#     offspring1 = column.CloneDescendant()
#     offspring2 = column.CloneDescendant()
#
#     for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
#         if differentia_width == 64:
#             assert (
#                 hstrat.calc_rank_of_last_retained_commonality_between(
#                     c1,
#                     c2,
#                 )
#                 == column.GetNumStrataDeposited() - 1
#             )
#             assert (
#                 hstrat.calc_rank_of_last_retained_commonality_between(
#                     c2,
#                     c1,
#                 )
#                 == column.GetNumStrataDeposited() - 1
#             )
#         elif differentia_width == 1:
#             assert (
#                 hstrat.calc_rank_of_last_retained_commonality_between(
#                     c1,
#                     c2,
#                 )
#                 < column.GetNumStrataDeposited() - 1
#             )
#             assert (
#                 hstrat.calc_rank_of_last_retained_commonality_between(
#                     c2,
#                     c1,
#                 )
#                 < column.GetNumStrataDeposited() - 1
#             )
#
#     for c in [column, offspring1, offspring2]:
#         assert hstrat.calc_rank_of_last_retained_commonality_between(
#             c, c, 0.8
#         ) == c.GetNumStrataDeposited() - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
#             significance_level=0.2,
#         )
#         assert hstrat.calc_rank_of_last_retained_commonality_between(
#             c, c, 0.99
#         ) == c.GetNumStrataDeposited() - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
#             significance_level=0.01,
#         )
#
#     for generation in range(100):
#         offspring1.DepositStratum()
#         offspring2.DepositStratum()
#
#     for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
#         if differentia_width == 64:
#             assert (
#                 hstrat.calc_rank_of_last_retained_commonality_between(
#                     c1,
#                     c2,
#                 )
#                 == column.GetNumStrataDeposited() - 1
#             )
#             assert (
#                 hstrat.calc_rank_of_last_retained_commonality_between(
#                     c2,
#                     c1,
#                 )
#                 == column.GetNumStrataDeposited() - 1
#             )
#         elif differentia_width == 1:
#             assert hstrat.calc_rank_of_last_retained_commonality_between(
#                 c1, c2, 0.999999
#             ) < hstrat.calc_rank_of_last_retained_commonality_between(
#                 c2, c1, 0.8
#             )
#
#     for c in [column, offspring1, offspring2]:
#         assert hstrat.calc_rank_of_last_retained_commonality_between(
#             c, c, 0.8
#         ) == c.GetNumStrataDeposited() - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
#             significance_level=0.2,
#         )
#         assert hstrat.calc_rank_of_last_retained_commonality_between(
#             c, c, 0.99
#         ) == c.GetNumStrataDeposited() - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
#             significance_level=0.01,
#         )
#
#
# @pytest.mark.parametrize(
#     "ordered_store",
#     [
#         hstrat.HereditaryStratumOrderedStoreDict,
#         hstrat.HereditaryStratumOrderedStoreList,
#         hstrat.HereditaryStratumOrderedStoreTree,
#     ],
# )
# def test_IterRanksOfRetainedCommonalityBetweenGenerich2(ordered_store):
#     column = hstrat.HereditaryStratigraphicColumn(
#         stratum_differentia_bit_width=1,
#     )
#
#     while True:
#         offspring1 = column.CloneDescendant()
#         offspring2 = column.CloneDescendant()
#
#         res = hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring1,
#             offspring2,
#         )
#         assert res is None
#         assert res == hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring2,
#             offspring1,
#         )
#         for gen in range(10):
#             offspring1.DepositStratum()
#             offspring2.DepositStratum()
#
#         # should happen about every 1 in 20 times
#         if (
#             hstrat.calc_rank_of_last_retained_commonality_between(
#                 offspring2,
#                 offspring1,
#             )
#             is not None
#         ):
#             break
#
#     column = hstrat.HereditaryStratigraphicColumn(
#         stratum_differentia_bit_width=1,
#     )
#
#     for __ in range(100):
#         column.DepositStratum()
#
#     for rep in range(5):
#         offspring1 = column.CloneDescendant()
#         offspring2 = column.CloneDescendant()
#
#         res = hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring1,
#             offspring2,
#         )
#         assert res is not None
#         assert res == hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring2,
#             offspring1,
#         )
#
#         for gen in range(10):
#             offspring1.DepositStratum()
#             offspring2.DepositStratum()
#
#         res = hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring1,
#             offspring2,
#         )
#         assert res is not None
#         assert res == hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring2,
#             offspring1,
#         )
#
#
# @pytest.mark.parametrize(
#     "ordered_store",
#     [
#         hstrat.HereditaryStratumOrderedStoreDict,
#         hstrat.HereditaryStratumOrderedStoreList,
#         hstrat.HereditaryStratumOrderedStoreTree,
#     ],
# )
# def test_IterRanksOfRetainedCommonalityBetweenGenerich3(ordered_store):
#     column = hstrat.HereditaryStratigraphicColumn(
#         stratum_differentia_bit_width=64,
#         stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
#     )
#
#     for generation in range(100):
#         column.DepositStratum()
#
#     offspring1 = column.CloneDescendant()
#     offspring2 = column.CloneDescendant()
#
#     for generation in range(100):
#         offspring1.DepositStratum()
#         offspring2.DepositStratum()
#
#     for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
#         assert c1.GetNumStrataRetained() == 2
#         assert c2.GetNumStrataRetained() == 2
#
#         assert (
#             hstrat.calc_rank_of_last_retained_commonality_between(
#                 c1,
#                 c2,
#             )
#             == 0
#         )
#         assert (
#             hstrat.calc_rank_of_last_retained_commonality_between(
#                 c2,
#                 c1,
#             )
#             == 0
#         )
#
#
# @pytest.mark.parametrize(
#     "ordered_store",
#     [
#         hstrat.HereditaryStratumOrderedStoreDict,
#         hstrat.HereditaryStratumOrderedStoreList,
#         hstrat.HereditaryStratumOrderedStoreTree,
#     ],
# )
# def test_IterRanksOfRetainedCommonalityBetweenGenerich4(ordered_store):
#     column = hstrat.HereditaryStratigraphicColumn(
#         stratum_differentia_bit_width=1,
#         stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
#     )
#
#     while True:
#         offspring1 = column.CloneDescendant()
#         offspring2 = column.CloneDescendant()
#
#         res = hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring1,
#             offspring2,
#         )
#         assert res is None
#         assert res == hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring2,
#             offspring1,
#         )
#         for gen in range(10):
#             offspring1.DepositStratum()
#             offspring2.DepositStratum()
#
#         # should happen about every 1 in 20 times
#         if (
#             hstrat.calc_rank_of_last_retained_commonality_between(
#                 offspring2,
#                 offspring1,
#             )
#             is not None
#         ):
#             break
#
#     column = hstrat.HereditaryStratigraphicColumn(
#         stratum_differentia_bit_width=1,
#         stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
#     )
#
#     for __ in range(100):
#         column.DepositStratum()
#
#     for rep in range(5):
#         offspring1 = column.CloneDescendant()
#         offspring2 = column.CloneDescendant()
#
#         res = hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring1,
#             offspring2,
#         )
#         assert res is not None
#         assert res == hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring2,
#             offspring1,
#         )
#
#         for gen in range(10):
#             offspring1.DepositStratum()
#             offspring2.DepositStratum()
#
#         res = hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring1,
#             offspring2,
#         )
#         assert res is not None
#         assert res == hstrat.calc_rank_of_last_retained_commonality_between(
#             offspring2,
#             offspring1,
#         )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8],
)
def test_IterRanksOfRetainedCommonalityBetweenGenerich5(
    ordered_store,
    differentia_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_width,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    assert column.GetNumStrataDeposited() == 101
    assert offspring1.GetNumStrataDeposited() == 102
    assert offspring2.GetNumStrataDeposited() == 102

    for c1, c2 in it.permutations([column, offspring1, offspring2], 2):
        if differentia_width == 64 or c1 is column or c2 is column:
            assert [
                *iter_ranks_of_retained_commonality_between(
                    c1,
                    c2,
                )
            ] == [*column.IterRetainedRanks()]
        else:
            assert all(
                rank >= column.GetNumStrataDeposited()
                and rank in c1.IterRetainedRanks()
                and rank in c2.IterRetainedRanks()
                for rank in (
                    set(
                        iter_ranks_of_retained_commonality_between(
                            c1,
                            c2,
                        )
                    )
                )
                - set(column.IterRetainedRanks())
            )

        assert all(
            rank in set(c1.IterRetainedRanks()) & set(c2.IterRetainedRanks())
            for rank in iter_ranks_of_retained_commonality_between(c2, c1)
        )
        assert all(
            (
                c1.GetStratumAtRank(rank).GetDifferentia()
                == c2.GetStratumAtRank(rank).GetDifferentia()
            )
            == (rank in iter_ranks_of_retained_commonality_between(c2, c1))
            for rank in (
                set(c1.IterRetainedRanks()) & set(c2.IterRetainedRanks())
            )
        )

    offspring1.DepositStrata(99)
    offspring2.DepositStrata(99)

    for c1, c2 in it.permutations([column, offspring1, offspring2], 2):
        if differentia_width == 64 or c1 is column or c2 is column:
            assert [
                *iter_ranks_of_retained_commonality_between(
                    c1,
                    c2,
                )
            ] == [*column.IterRetainedRanks()]
        else:
            assert all(
                rank >= column.GetNumStrataDeposited()
                and rank in c1.IterRetainedRanks()
                and rank in c2.IterRetainedRanks()
                for rank in (
                    set(
                        iter_ranks_of_retained_commonality_between(
                            c1,
                            c2,
                        )
                    )
                )
                - set(column.IterRetainedRanks())
            )

        assert all(
            rank in set(c1.IterRetainedRanks()) & set(c2.IterRetainedRanks())
            for rank in iter_ranks_of_retained_commonality_between(c2, c1)
        )
        assert all(
            all(
                c1.GetStratumAtRank(r).GetDifferentia()
                == c2.GetStratumAtRank(r).GetDifferentia()
                for r in (
                    set(c1.IterRetainedRanks()) & set(c2.IterRetainedRanks())
                )
                if r <= rank
            )
            == (rank in iter_ranks_of_retained_commonality_between(c2, c1))
            for rank in (
                set(c1.IterRetainedRanks()) & set(c2.IterRetainedRanks())
            )
        )



import numbers
import typing

from iterpop import iterpop as ip
import networkx as nx

from ._find_chronological_roots import find_chronological_roots


def find_chronological_root(
    tree: nx.Graph,
    leaf_node_origin_times: typing.Dict[typing.Any, numbers.Number],
) -> str:
    """Convenience wrapper for `find_chronological_roots`, assuming phylogeny descends from a single common ancestor."""
    return ip.popsingleton(
        find_chronological_roots(tree, leaf_node_origin_times)
    )



from copy import deepcopy
import pickle
import tempfile
import unittest

from hstrat import hstrat


class TestHereditaryStratigraphicColumnBundle(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_equality(self):
        def make_populated_bundle():
            return hstrat.HereditaryStratigraphicColumnBundle(
                {
                    "test": hstrat.HereditaryStratigraphicColumn(),
                    "control": hstrat.HereditaryStratigraphicColumn(
                        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
                    ),
                }
            )

        assert make_populated_bundle() != make_populated_bundle()
        column1 = make_populated_bundle()
        column2 = column1
        assert column1 == column2
        assert column1 == deepcopy(column2)

        column3 = make_populated_bundle()
        assert column3 != column2

    def test_clone(self):
        def make_populated_bundle():
            return hstrat.HereditaryStratigraphicColumnBundle(
                {
                    "test": hstrat.HereditaryStratigraphicColumn(),
                    "control": hstrat.HereditaryStratigraphicColumn(
                        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
                    ),
                }
            )

        column1 = make_populated_bundle()
        column2 = column1.Clone()
        column3 = column1.Clone()

        assert column1 == column2
        assert column2 == column3
        assert column1 == column3

        column1.DepositStratum()

        assert column1 != column2
        assert column2 == column3
        assert column1 != column3

    def test_pickle(self):
        def make_populated_bundle():
            return hstrat.HereditaryStratigraphicColumnBundle(
                {
                    "test": hstrat.HereditaryStratigraphicColumn(),
                    "control": hstrat.HereditaryStratigraphicColumn(
                        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
                    ),
                }
            )

        column1 = make_populated_bundle()
        column2 = column1.Clone()
        column3 = column1.CloneDescendant()

        column3.DepositStratum()

        for __ in range(100):
            column1.DepositStratum()

        for original in column1, column2, column3:
            with tempfile.TemporaryDirectory() as tmp_path:
                with open(f"{tmp_path}/data", "wb") as tmp_file:
                    pickle.dump(original, tmp_file)

                with open(f"{tmp_path}/data", "rb") as tmp_file:
                    reconstituted = pickle.load(tmp_file)
                    assert reconstituted == original

    def test_forwarding_fallback(self):
        bundle1 = hstrat.HereditaryStratigraphicColumnBundle(
            {
                "test": hstrat.HereditaryStratigraphicColumn(
                    stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
                    stratum_differentia_bit_width=1,
                ),
                "control": hstrat.HereditaryStratigraphicColumn(
                    stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
                ),
            }
        )

        for __ in range(100):
            bundle1.DepositStratum()
        bundle2 = bundle1.Clone()
        for __ in range(100):
            bundle1.DepositStratum()
        for __ in range(100):
            bundle2.DepositStratum()

        assert (
            hstrat.does_have_any_common_ancestor(
                bundle1["test"], bundle2["test"]
            )
            is None
        )
        assert (
            hstrat.does_have_any_common_ancestor(
                bundle1["control"], bundle2["control"]
            )
            is True
        )

        assert (
            hstrat.does_have_any_common_ancestor(
                bundle1["test"], bundle2["test"], confidence_level=0.49
            )
            is True
        )
        assert (
            hstrat.does_have_any_common_ancestor(
                bundle1["control"], bundle2["control"], confidence_level=0.49
            )
            is True
        )

        res = bundle1.GetNumStrataRetained()
        assert (
            0 < res["test"] < res["control"] <= bundle1.GetNumStrataDeposited()
        )

        assert bundle1._stratum_differentia_bit_width == {
            "test": 1,
            "control": 64,
        }


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pytest

from hstrat.hstrat import recency_proportional_resolution_curbed_algo
from hstrat.stratum_retention_strategy.stratum_retention_algorithms._impl import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce,
)


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        np.random.default_rng(1).integers(
            10**2,
            size=10**2,
        ),
        np.random.default_rng(1).integers(
            10**3,
            size=10**2,
        ),
    ],
)
def test_policy_consistency(size_curb, time_sequence):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
        spec,
    )
    for num_strata_deposited_a in time_sequence:
        for num_strata_deposited_b in (
            num_strata_deposited_a // 3,
            num_strata_deposited_a // 2,
            num_strata_deposited_a,
            num_strata_deposited_a + 1,
            num_strata_deposited_a + 10,
            num_strata_deposited_a + 100,
        ):
            if 0 in (num_strata_deposited_a, num_strata_deposited_b):
                continue
            policy_requirement = policy.CalcMrcaUncertaintyRelUpperBound(
                num_strata_deposited_a,
                num_strata_deposited_b,
                CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce()(
                    policy,
                    num_strata_deposited_a,
                    num_strata_deposited_b,
                ),
            )
            for which in (
                instance,
                recency_proportional_resolution_curbed_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
                    spec
                ),
            ):
                assert (
                    which(
                        policy,
                        num_strata_deposited_a,
                        num_strata_deposited_b,
                    )
                    == policy_requirement
                )


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(size_curb):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
        spec
    )

    assert instance == instance
    assert (
        instance
        == recency_proportional_resolution_curbed_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
            spec,
        )
    )
    assert instance is not None



import os

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_assign_contiguous_ids,
    alifestd_find_leaf_ids,
    alifestd_prune_extinct_lineages_asexual,
    alifestd_to_working_format,
    alifestd_topological_sort,
    alifestd_try_add_ancestor_id_col,
    alifestd_unfurl_lineage_asexual,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_assign_contiguous_ids,
        alifestd_to_working_format,
        alifestd_topological_sort,
        alifestd_try_add_ancestor_id_col,
        lambda x: x.sample(frac=1, random_state=1),
        lambda x: x,
    ],
)
def test_alifestd_prune_extinct_lineages_asexual_destruction_time_nop(
    phylogeny_df, apply
):

    assert "destruction_time" in phylogeny_df
    leaf_mask = phylogeny_df["id"].isin(
        {*alifestd_find_leaf_ids(phylogeny_df)}
    )
    leaf_destruction_times = phylogeny_df.loc[leaf_mask, "destruction_time"]
    assert (
        leaf_destruction_times.isna() | np.isinf(leaf_destruction_times)
    ).all()
    # because the source dataframes are pruned, pruning will be a nop
    phylogeny_df = apply(phylogeny_df.copy()).reset_index(drop=True)

    phylogeny_df_ = phylogeny_df.copy()
    pruned_df = alifestd_prune_extinct_lineages_asexual(phylogeny_df)
    assert phylogeny_df.equals(phylogeny_df_)

    assert len(phylogeny_df) == len(pruned_df)

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)
    assert {*phylogeny_df} == {*pruned_df}
    assert phylogeny_df.equals(pruned_df), phylogeny_df.compare(pruned_df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_assign_contiguous_ids,
        alifestd_to_working_format,
        alifestd_topological_sort,
        alifestd_try_add_ancestor_id_col,
        lambda x: x.sample(frac=1, random_state=1),
        lambda x: x,
    ],
)
def test_alifestd_prune_extinct_lineages_asexual(phylogeny_df, apply):
    # because the source dataframes are pruned, pruning will be a nop
    phylogeny_df = apply(phylogeny_df.copy())

    extant_mask = np.random.choice([True, False], size=len(phylogeny_df))
    phylogeny_df["extant"] = extant_mask

    phylogeny_df_ = phylogeny_df.copy()
    pruned_df = alifestd_prune_extinct_lineages_asexual(phylogeny_df)
    assert len(pruned_df) < len(phylogeny_df)
    assert set(pruned_df["id"]) >= set(phylogeny_df.loc[extant_mask, "id"])

    assert alifestd_validate(pruned_df)
    assert phylogeny_df.equals(phylogeny_df_)

    for leaf_id in alifestd_find_leaf_ids(pruned_df):
        assert [*alifestd_unfurl_lineage_asexual(phylogeny_df, leaf_id)] == [
            *alifestd_unfurl_lineage_asexual(pruned_df, leaf_id)
        ]


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_assign_contiguous_ids,
        alifestd_to_working_format,
        alifestd_topological_sort,
        alifestd_try_add_ancestor_id_col,
        lambda x: x.sample(frac=1, random_state=1),
        lambda x: x,
    ],
)
def test_alifestd_prune_extinct_lineages_asexual_independent_trees(
    phylogeny_df, apply
):
    # because the source dataframes are pruned, pruning will be a nop
    phylogeny_df = apply(phylogeny_df.copy())
    phylogeny_df["extant"] = False

    first_df = phylogeny_df.copy()
    extant_mask = first_df["id"].isin(alifestd_find_leaf_ids(first_df))
    first_df.loc[extant_mask, "extant"] = True

    second_df = phylogeny_df.copy()

    pruned_df = alifestd_prune_extinct_lineages_asexual(
        alifestd_aggregate_phylogenies([first_df, second_df]),
    )
    assert len(pruned_df) == len(phylogeny_df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_assign_contiguous_ids,
        alifestd_to_working_format,
        alifestd_topological_sort,
        alifestd_try_add_ancestor_id_col,
        lambda x: x.sample(frac=1, random_state=1),
        lambda x: x,
    ],
)
def test_alifestd_prune_extinct_lineages_asexual_ambiguous_extant(
    phylogeny_df, apply
):
    # because the source dataframes are pruned, pruning will be a nop
    phylogeny_df = apply(phylogeny_df.copy())
    phylogeny_df.drop("destruction_time", axis="columns", inplace=True)

    with pytest.raises(ValueError):
        alifestd_prune_extinct_lineages_asexual(phylogeny_df)



import pytest

from hstrat._auxiliary_lib import argsort


@pytest.fixture
def seq():
    return [4.0, 2.0, 1.0, 3.0]


def test_argsort(seq):
    expected_output = [2, 1, 3, 0]
    assert argsort(seq) == expected_output


def test_argsort_reverse(seq):
    expected_output = [0, 3, 1, 2]
    assert argsort(seq, reverse=True) == expected_output


def test_argsort_empty_input():
    assert argsort([]) == []


def test_argsort_single_input():
    assert argsort([1]) == [0]



from copy import copy
import operator
import typing

from ._HereditaryStratigraphicColumn import HereditaryStratigraphicColumn


class HereditaryStratigraphicColumnBundle:
    """Packages multiple HereditaryStratigraphicColumn instances together.

    Allows packaged columsn to conveniently advance in sync along a line of
    descent with a similar interface to an individual
    HereditaryStratigraphicColumn.
    """

    __slots__ = ("_columns",)

    _columns: typing.Dict[str, HereditaryStratigraphicColumn]

    def __init__(
        self: "HereditaryStratigraphicColumnBundle",
        columns: typing.Dict[str, HereditaryStratigraphicColumn],
    ):
        """Construct bundle.

        Parameters
        ----------
        columns : dict
            HereditaryStratigraphicColumn objects to bundle together, each
            associated with a unique names as its key.
        """
        assert len(columns), (
            "Must provide at least one column to "
            "HereditaryStratigraphicColumnBundle."
        )
        assert (
            len({c.GetNumStrataDeposited() for c in columns.values()}) == 1
        ), (
            "All columns provided ot HereditaryStratigraphicColumnBundle "
            "must have same number strata deposited. "
        )
        self._columns = columns

    def __iter__(
        self: "HereditaryStratigraphicColumnBundle",
    ) -> typing.Iterator[str]:
        """Iterate over held columns."""
        yield from self._columns

    def __getitem__(
        self: "HereditaryStratigraphicColumnBundle",
        key: str,
    ) -> HereditaryStratigraphicColumn:
        """Brackets operator; access constituent column by name."""
        return self._columns[key]

    def __eq__(
        self: "HereditaryStratigraphicColumnBundle",
        other: "HereditaryStratigraphicColumnBundle",
    ) -> bool:
        """Compare for value-wise equality."""
        # adapted from https://stackoverflow.com/a/4522896
        return (
            isinstance(
                other,
                self.__class__,
            )
            and self.__slots__ == other.__slots__
            and all(
                getter(self) == getter(other)
                for getter in [
                    operator.attrgetter(attr) for attr in self.__slots__
                ]
            )
        )

    def DepositStratum(
        self: "HereditaryStratigraphicColumnBundle",
        annotation: typing.Optional[typing.Any] = None,
    ) -> None:
        """Elapse a generation on all constituent columns.

        Parameters
        ----------
        annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum's generation in its
            line of descent.
        """
        for column in self._columns.values():
            column.DepositStratum(
                annotation=annotation,
            )

    def GetNumStrataDeposited(
        self: "HereditaryStratigraphicColumnBundle",
    ) -> int:
        """How many strata have been deposited on constituent columns?

        Should be identical across constituent columns.
        """
        return next(iter(self._columns.values())).GetNumStrataDeposited()

    def Clone(
        self: "HereditaryStratigraphicColumnBundle",
    ) -> "HereditaryStratigraphicColumnBundle":
        """Create a copy of the bundle.

        Copy contains identical data but may be freely altered without
        affecting data within this bundle.
        """
        # shallow copy
        result = copy(self)
        # do semi-shallow clone on select elements
        # see https://stackoverflow.com/a/5861653 for performance consierations
        result._columns = {k: v.Clone() for k, v in self._columns.items()}
        return result

    def CloneDescendant(
        self: "HereditaryStratigraphicColumnBundle",
        stratum_annotation: typing.Optional[typing.Any] = None,
    ) -> "HereditaryStratigraphicColumnBundle":
        """Return a cloned bundle that has had an additional stratum deposited.

        Does not alter self.

        Parameters
        ----------
        stratum_annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum deposition in the
            line of descent.
        """
        res = self.Clone()
        res.DepositStratum(annotation=stratum_annotation)
        return res

    def __getattr__(
        self: "HereditaryStratigraphicColumnBundle",
        attr: str,
    ) -> typing.Union[typing.Callable, typing.Dict]:
        """Forward all unknown method calls and property accesses.

        Forward to underlying ereditaryStratigraphicColumns, returning a dict
        of results for each column key stored.

        Note that __getattr__ is only called after other attribute lookup (i.e.,
        explicitly provided methods and properties) has failed.
        """
        # Adapted from
        # https://rosettacode.org/wiki/Respond_to_an_unknown_method_call#Python

        # raise AttributeError for dunder methods so that callers expecting
        # them unimplemented can catch and run their fallbacks
        if "__" in attr:
            raise AttributeError

        def arg_debundler(args, column_name) -> typing.List:
            """If any args are column bundles, extract the focal column."""
            return [
                arg
                if not isinstance(arg, self.__class__)
                else arg[column_name]
                for arg in args
            ]

        def kwarg_debundler(kwargs, column_name) -> typing.Dict:
            """If any kwarg vals are column bundles, extract focal column."""
            return {
                k: v if not isinstance(v, self.__class__) else v[column_name]
                for k, v in kwargs.items()
            }

        if any(
            callable(getattr(column, attr))
            for column in self._columns.values()
        ):
            # method forwarding
            assert all(
                callable(getattr(column, attr))
                for column in self._columns.values()
            )

            def forwarded(*args, **kwargs) -> typing.Dict:
                """Apply method to each column independently.

                Extracts the corresponding column from any column bundles passed as arguments. Returns a dict mapping each column name to its result.
                """
                return {
                    column_name: getattr(column, attr)(
                        *arg_debundler(args, column_name),
                        **kwarg_debundler(kwargs, column_name),
                    )
                    for column_name, column in self._columns.items()
                }

            return forwarded

        else:
            # property forwarding
            return {
                column_name: getattr(column, attr)
                for column_name, column in self._columns.items()
            }



import numpy as np

from ...._auxiliary_lib import count_unique, indices_of_unique, is_in_unit_test
from ...perfect_tracking import GarbageCollectingPhyloTracker


def _apply_island_migrations(
    pop_arr: np.array,
    pop_tracker: GarbageCollectingPhyloTracker,
    num_niches: int,
    island_size: int,
    island_niche_size: int,
    p_island_migration: float,
) -> None:
    """Copy organisms to neighboring islands, altering `pop_arr` inplace.

    See `evolve_fitness_trait_population` for parameter specifications.
    """

    num_island_migrations = np.random.binomial(
        n=len(pop_arr),
        p=p_island_migration,
    )
    # end early for no-op
    if island_size == len(pop_arr) or not num_island_migrations:
        return

    copyto_locs = np.empty(num_island_migrations, dtype=np.int32)
    copyfrom_locs = np.empty(num_island_migrations, dtype=np.int32)
    num_finalized = 0  # loop until all copyto_locs are distinct
    while num_finalized < num_island_migrations:
        copyfrom_locs[num_finalized:] = np.random.randint(
            len(pop_arr), size=num_island_migrations - num_finalized
        )
        copyto_signs = (
            np.random.randint(2, size=num_island_migrations - num_finalized)
            * 2
            - 1
        )
        copyto_locs[num_finalized:] = (
            copyfrom_locs[num_finalized:] + copyto_signs * island_size
        )
        # wrap overflow
        copyto_locs[copyto_locs >= len(pop_arr)] -= len(pop_arr)

        # move unique to front
        indices_of_unique_ = indices_of_unique(copyto_locs)
        num_finalized = len(indices_of_unique_)
        copyfrom_locs[:num_finalized] = copyfrom_locs[indices_of_unique_]
        copyto_locs[:num_finalized] = copyto_locs[indices_of_unique_]

    if is_in_unit_test():
        assert len(copyto_locs) == len(copyfrom_locs)
        assert all(
            copyto_locs // island_niche_size % num_niches
            == copyfrom_locs // island_niche_size % num_niches
        )
        assert np.all(
            copyto_locs // island_size != copyfrom_locs // island_size
        )
        assert count_unique(copyto_locs) == len(copyto_locs)

    pop_arr[copyto_locs] = pop_arr[copyfrom_locs].copy()
    pop_tracker.ApplyLocPasteovers(copyfrom_locs, copyto_locs)



import sys


# adapted from https://medium.com/analytics-vidhya/increase-maximum-recursion-depth-in-python-using-context-manager-1c67eaf4e71b
class RecursionLimit:
    """Context manager to temporarily override recursion limit."""

    def __init__(self, limit):
        self.limit = limit
        self.cur_limit = sys.getrecursionlimit()

    def __enter__(self):
        sys.setrecursionlimit(self.limit)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        sys.setrecursionlimit(self.cur_limit)



import numbers

from ..._auxiliary_lib import cmp_approx


class BubbleWrappedPrior:
    """Asserts that wrapped prior receives valid inputs and produces valid
    output."""

    _prior: object

    def __init__(self: "BubbleWrappedPrior", wrapee: object):
        self._prior = wrapee

    def CalcIntervalProbabilityProxy(
        self: "BubbleWrappedPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Characterize the prior probability of the MRCA generation falling
        within an interval range.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The proxy statistic, proportional to the true estimated interval
            probability of the MRCA value by a fixed (but unspecified) constant
            proportion.
        """
        assert isinstance(begin_rank, numbers.Integral)
        assert isinstance(end_rank, numbers.Integral)
        assert 0 <= begin_rank < end_rank
        res = self._prior.CalcIntervalProbabilityProxy(begin_rank, end_rank)
        return res

    def CalcIntervalConditionedMean(
        self: "BubbleWrappedPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Calcualate the centriod of prior probability mass within an interval
        of possible MRCA generations.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The prior expected generation of MRCA conditioned on the assumption
            that the MRCA falls within the given interval.
        """
        assert isinstance(begin_rank, numbers.Integral)
        assert isinstance(end_rank, numbers.Integral)
        assert 0 <= begin_rank < end_rank
        res = self._prior.CalcIntervalConditionedMean(begin_rank, end_rank)
        assert cmp_approx(begin_rank, res) <= 0
        assert res < end_rank
        return res



import numpy as np

from ...._auxiliary_lib import jit, jit_numpy_bool_t


# implemented as free function (not member) so self param doesn't interfere
# with nopython directive
@jit(nopython=True)
def _discern_referenced_rows(
    parentage_buffer: np.array,
    num_records: int,
    population_size: int,
    below_row: int = 0,
) -> np.array:
    """Return array of indices for all rows at or below `below_row` that
    represent organisms with extant lineages.

    Indices of rows with extinct lineages are not included.
    """

    assert below_row >= 0
    assert num_records >= below_row

    referenced_rows = np.zeros(num_records - below_row, dtype=jit_numpy_bool_t)
    for pop_position in range(population_size):
        idx = num_records - population_size + pop_position
        while (
            idx != parentage_buffer[idx]
            and idx >= below_row
            and not referenced_rows[idx - below_row]
        ):
            referenced_rows[idx - below_row] = True
            idx = parentage_buffer[idx]

        if idx >= below_row:
            referenced_rows[idx - below_row] = True

    return referenced_rows



# copied from https://github.com/mmore500/alifedata-phyloinformatics-convert/blob/9287db7c101604cf49df16617fb32a632122fbd9/alifedata_phyloinformatics_convert/anytree_tree_to_alife_dataframe.py
# with recursive iterators swapped for iterative implementatins


import anytree
from nanto import isanan
import opytional as opyt
import pandas as pd

from ._AnyTreeFastLevelOrderIter import AnyTreeFastLevelOrderIter
from ._AnyTreeFastPreOrderIter import AnyTreeFastPreOrderIter


def anytree_tree_to_alife_dataframe(
    tree: anytree.AnyNode,
) -> pd.DataFrame:
    """Convert a anytree tree to a dataframe formatted to the
    artificial life community data format standards.

    The following Node object attributes will automatically be exported to
    dataframe columns, if available:
        * edge_length,
        * id,
        * label,
        * name,
        * destruction_time,
        * origin_time, and
        * taxon_label.

    Parameters
    ----------
    tree:
        anytree tree to convert.
    """

    # set up node origin times if any edge lengths set
    if any(
        getattr(node, "edge_length", None) is not None
        for node in AnyTreeFastPreOrderIter(tree)
    ):
        if not hasattr(tree, "origin_time"):
            tree.origin_time = opyt.or_value(
                getattr(tree, "edge_length", None), 0
            )
        else:
            assert tree.origin_time is not None
        for node in AnyTreeFastLevelOrderIter(tree):
            parent = node.parent
            if parent is not None and not hasattr(node, "origin_time"):
                if None not in (
                    getattr(parent, "origin_time", None),
                    getattr(node, "edge_length", None),
                ) and (
                    not isanan(parent.origin_time)
                    and not isanan(node.edge_length)
                ):
                    node.origin_time = parent.origin_time + node.edge_length

    # attach ids to nodes, if needed
    for fallback_id, node in enumerate(AnyTreeFastLevelOrderIter(tree)):
        if not hasattr(node, "id"):
            node.id = fallback_id
        else:
            assert isinstance(node.id, int)

    return pd.DataFrame.from_records(
        [
            {
                "id": node.id,
                "ancestor_list": f"[{opyt.apply_if(node.parent, lambda x: x.id)}]",
                "edge_length": getattr(node, "edge_length", None),
                "label": getattr(node, "label", None),
                "name": getattr(node, "name", None),
                "origin_time": getattr(node, "origin_time", None),
                "destruction_time": getattr(node, "destruction_time", None),
                "taxon_label": getattr(node, "taxon_label", None),
            }
            for node in AnyTreeFastLevelOrderIter(tree)
        ]
    ).dropna(axis=1, how="all")



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_leaf_ids,
    alifestd_make_empty,
    alifestd_mark_leaves,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    original = phylogeny_df.copy()

    result = alifestd_mark_leaves(phylogeny_df)

    assert alifestd_validate(result)
    assert original.equals(phylogeny_df)

    leaf_ids = [*alifestd_find_leaf_ids(phylogeny_df)]
    for leaf_id in leaf_ids:
        is_leaf = result[result["id"] == leaf_id]["is_leaf"].squeeze()
        assert is_leaf

    assert all(0 <= result["is_leaf"])
    assert all(result["is_leaf"] <= 1)


def test_empty():
    res = alifestd_mark_leaves(alifestd_make_empty())
    assert "is_leaf" in res
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_leaves(
        phylogeny_df,
        mutate=mutate,
    )
    assert not result_df.loc[0, "is_leaf"]
    assert not result_df.loc[1, "is_leaf"]
    assert result_df.loc[2, "is_leaf"]

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[0]", "[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_leaves(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[1, "is_leaf"]
    assert not result_df.loc[0, "is_leaf"]
    assert result_df.loc[2, "is_leaf"]
    assert result_df.loc[3, "is_leaf"]

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_leaves(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[1, "is_leaf"]
    assert not result_df.loc[0, "is_leaf"]
    assert result_df.loc[2, "is_leaf"]
    assert result_df.loc[3, "is_leaf"]

    if not mutate:
        assert original_df.equals(phylogeny_df)



import os

import alifedata_phyloinformatics_convert as apc
import numpy as np
import pytest
from tqdm import tqdm

from hstrat import hstrat
from hstrat._auxiliary_lib import omit_last

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "num_islands",
    [
        1,
        4,
    ],
)
@pytest.mark.parametrize(
    "num_niches",
    [
        1,
        4,
    ],
)
@pytest.mark.parametrize(
    "num_generations",
    [
        10,
        200,
    ],
)
@pytest.mark.parametrize(
    "population_size",
    [
        128,
        256,
    ],
)
def test_evolve_fitness_trait_population(
    num_islands,
    num_niches,
    num_generations,
    population_size,
):
    alife_df = hstrat.evolve_fitness_trait_population(
        num_islands=num_islands,
        num_niches=num_niches,
        num_generations=num_generations,
        population_size=population_size,
        tournament_size=2,
        progress_wrap=tqdm,
    )

    assert "trait" in alife_df
    assert "loc" in alife_df
    assert "island" in alife_df
    assert "niche" in alife_df

    assert all(0 <= island <= num_islands for island in alife_df["island"])
    assert all(0 <= niche <= num_niches for niche in alife_df["niche"])
    assert num_islands * num_niches == len(
        set(
            (island, niche)
            for __, (island, niche) in alife_df[["island", "niche"]].iterrows()
        )
    )

    tree = apc.alife_dataframe_to_dendropy_tree(alife_df)

    assert len(tree) == population_size
    assert len(set(leaf_node.level() for leaf_node in tree.leaf_node_iter()))
    assert all(
        leaf_node.level() == num_generations + 1
        for leaf_node in tree.leaf_node_iter()
    )


@pytest.mark.parametrize(
    "num_islands",
    [
        1,
        4,
    ],
)
@pytest.mark.parametrize(
    "num_niches",
    [
        1,
        4,
    ],
)
@pytest.mark.parametrize(
    "p_island_migration",
    [
        0.0,
        0.1,
        1.0,
    ],
)
@pytest.mark.parametrize(
    "p_niche_invasion",
    [
        0.0,
        0.1,
        1.0,
    ],
)
@pytest.mark.parametrize(
    "mut_distn",
    [
        np.random.standard_normal,
        np.random.laplace,
        lambda size: np.zeros(shape=size),
    ],
)
def test_evolve_fitness_trait_population_swaps(
    num_islands,
    num_niches,
    p_island_migration,
    p_niche_invasion,
    mut_distn,
):
    alife_df = hstrat.evolve_fitness_trait_population(
        num_islands=num_islands,
        num_niches=num_niches,
        num_generations=200,
        population_size=128,
        p_island_migration=p_island_migration,
        p_niche_invasion=p_niche_invasion,
        mut_distn=mut_distn,
        progress_wrap=tqdm,
    )

    assert "trait" in alife_df
    assert "loc" in alife_df
    assert "island" in alife_df
    assert "niche" in alife_df

    tree = apc.alife_dataframe_to_dendropy_tree(
        alife_df,
        setattrs=("island", "niche"),
    )

    assert any(
        len(
            set(
                ancestor.island
                for ancestor in omit_last(leaf.ancestor_iter(inclusive=True))
            )
        )
        > 1
        for leaf in tree.leaf_node_iter()
    ) == (p_island_migration and num_islands > 1)
    assert any(
        len(
            set(
                ancestor.niche
                for ancestor in omit_last(leaf.ancestor_iter(inclusive=True))
            )
        )
        > 1
        for leaf in tree.leaf_node_iter()
    ) == (p_niche_invasion and num_niches > 1)


def test_evolve_fitness_trait_population_selection():
    drift_alife_df = hstrat.evolve_fitness_trait_population(
        num_generations=200,
        population_size=128,
        tournament_size=1,
        progress_wrap=tqdm,
    )
    drift_tree = apc.alife_dataframe_to_dendropy_tree(
        drift_alife_df,
        setattrs=("trait",),
    )
    drift_tree_mean_fitness = np.mean(
        [node.trait for node in drift_tree.leaf_node_iter()]
    )

    weak_alife_df = hstrat.evolve_fitness_trait_population(
        num_generations=200,
        population_size=128,
        tournament_size=2,
        progress_wrap=tqdm,
    )
    weak_tree = apc.alife_dataframe_to_dendropy_tree(
        weak_alife_df,
        setattrs=("trait",),
    )
    weak_tree_mean_fitness = np.mean(
        [node.trait for node in weak_tree.leaf_node_iter()]
    )

    weak_bigpop_alife_df = hstrat.evolve_fitness_trait_population(
        num_generations=200,
        population_size=1024,
        tournament_size=2,
        progress_wrap=tqdm,
    )
    weak_bigpop_tree = apc.alife_dataframe_to_dendropy_tree(
        weak_bigpop_alife_df,
        setattrs=("trait",),
    )
    weak_bigpop_tree_mean_fitness = np.mean(
        [node.trait for node in weak_bigpop_tree.leaf_node_iter()]
    )

    strong_alife_df = hstrat.evolve_fitness_trait_population(
        num_generations=200,
        population_size=128,
        tournament_size=7,
        progress_wrap=tqdm,
    )
    strong_tree = apc.alife_dataframe_to_dendropy_tree(
        strong_alife_df,
        setattrs=("trait",),
    )
    strong_tree_mean_fitness = np.mean(
        [node.trait for node in strong_tree.leaf_node_iter()]
    )

    assert (
        drift_tree_mean_fitness
        < weak_tree_mean_fitness
        < strong_tree_mean_fitness
    )

    assert weak_tree_mean_fitness < weak_bigpop_tree_mean_fitness

    assert sum(1 for __ in drift_tree) > sum(1 for __ in strong_tree)


@pytest.mark.parametrize(
    "num_generations",
    [
        100,
        200,
    ],
)
@pytest.mark.parametrize(
    "population_size",
    [
        128,
        256,
    ],
)
@pytest.mark.parametrize(
    "share_common_ancestor",
    [
        True,
        False,
    ],
)
def test_evolve_fitness_trait_population_iter_epochs(
    num_generations,
    population_size,
    share_common_ancestor,
):

    epoch_iter = hstrat.evolve_fitness_trait_population(
        iter_epochs=True,
        num_generations=num_generations,
        population_size=population_size,
        share_common_ancestor=share_common_ancestor,
        progress_wrap=tqdm,
    )

    dfs = [next(epoch_iter) for __ in range(3)]

    series_sets = [
        set(
            # id and ancestor_list (ancestor id) depend on position within
            # extant organisms, so aren't conserved
            df.drop(["id", "ancestor_list"], axis=1).apply(
                tuple,
                axis=1,
            )
        )
        for df in dfs
    ]

    assert len(series_sets[0] & series_sets[1]) > num_generations - 1
    assert len(series_sets[0] & series_sets[2]) > num_generations * 2 - 1

    for epoch, df in enumerate(dfs):
        assert all(
            leaf_node.level()
            == (epoch + 1) * num_generations + share_common_ancestor
            for tree in apc.alife_dataframe_to_dendropy_trees(df)
            for leaf_node in tree.leaf_node_iter()
        )



import numpy as np
import pandas as pd


def get_nullable_mask(series: pd.Series) -> np.ndarray:
    """Get a boolean mask indicating which elements in a Pandas Series with
    nullable integer dtype are null.

    Parameters
    ----------
    series : pd.Series
        The Pandas Series to get the null mask for.

    Returns
    -------
    np.ndarray
        A 1-dimensional NumPy ndarray of boolean values. The i-th element of the
        array is True if the i-th element of the input series is null, and False
        otherwise.

    Notes
    -----
    This function returns the underlying boolean mask used by the Pandas Series
    object to represent null values. This mask is a direct view into the Series
    data, so no copy is made. Changes to the mask will propagate to the Series
    object, and vice versa.
    """
    return series._data.array._mask



import typing

import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_mrca_bounds_between import calc_rank_of_mrca_bounds_between


def calc_patristic_distance_bounds_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    prior: str,
    confidence_level: float = 0.95,
) -> typing.Optional[typing.Tuple[int, int]]:
    """What is the total phylogenetic distance along the branch path connecting
    first and second?

    Calculate confidence interval for patristic distance estimate. Branch
    length here is in terms of number of generations elapsed. So the
    calculated distance estimates the sum of the number of generations elapsed

    Parameters
    ----------
    prior : {"arbitrary"}
        Prior probability density distribution over possible generations of the
        MRCA.

        Currently only "arbitrary" supported.
    confidence_level : float, optional
        Bounds must capture what probability of containing the true patristic
        distance? Default 0.95.

    Returns
    -------
    (int, int), optional
        Inclusive lower and then exclusive upper bound on patristic distance
        estimate or None if no common ancestor between first and second can be
        resolved with sufficient confidence. (Sufficient confidence depends on
        confidence_level.)

    See Also
    --------
    calc_rank_of_earliest_detectable_mrca_between :
        Could any MRCA be detected between first and second? What is the rank
        of the earliest MRCA that could be reliably detected?
    does_definitively_have_no_common_anestor :
        Does the hereditary stratigraphic record definitively prove that first and second could not possibly share a common ancestor?

    Notes
    -----
    The true patristic distance is guaranteed to never fall below the returned bounds but may fall above.
    """

    mrca_rank_bounds = calc_rank_of_mrca_bounds_between(
        first, second, prior=prior
    )
    inclusive_exclusive_ub_lb_switch_correction = 2
    max_patristic_distance = (
        first.GetNumStrataDeposited() - 1 + second.GetNumStrataDeposited() - 1
    ) + inclusive_exclusive_ub_lb_switch_correction

    return opyt.apply_if(
        mrca_rank_bounds,
        lambda bounds: tuple(
            max_patristic_distance - 2 * bound for bound in reversed(bounds)
        ),
    )



"""Provides numba bool type or numpy bool type.

Type must be declared outside jit'ed function or numba fails.
"""

import numpy as np

from ._is_in_coverage_run import is_in_coverage_run

try:
    import numba as nb
except (ImportError, ModuleNotFoundError):
    jit_numpy_bool_t = np.bool_
else:
    if is_in_coverage_run():
        # when numba disabled for coverage testing,
        # using nb.types.bool_ causes numpy TypeError
        jit_numpy_bool_t = np.bool_
    else:  # pragma: no cover
        jit_numpy_bool_t = nb.types.bool_



"""Tools to extant populations of hereditary stratigraph columns corresponding
to a template phylogeny."""

from ._descend_template_phylogeny import descend_template_phylogeny
from ._descend_template_phylogeny_alifestd import (
    descend_template_phylogeny_alifestd,
)
from ._descend_template_phylogeny_biopython import (
    descend_template_phylogeny_biopython,
)
from ._descend_template_phylogeny_dendropy import (
    descend_template_phylogeny_dendropy,
)
from ._descend_template_phylogeny_naive import descend_template_phylogeny_naive
from ._descend_template_phylogeny_networkx import (
    descend_template_phylogeny_networkx,
)
from ._descend_template_phylogeny_posthoc import (
    descend_template_phylogeny_posthoc,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "descend_template_phylogeny",
    "descend_template_phylogeny_alifestd",
    "descend_template_phylogeny_biopython",
    "descend_template_phylogeny_dendropy",
    "descend_template_phylogeny_naive",
    "descend_template_phylogeny_networkx",
    "descend_template_phylogeny_posthoc",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



from ..._impl import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank as CalcMrcaUncertaintyRelUpperBoundAtPessimalRank_,
)


# must inherit rather than assign due to failure of attribute lookup when
# pickling otherwise
class CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank_
):
    pass



import math
import typing

from ...._auxiliary_lib import (
    AnyTreeFastPreOrderIter,
    anytree_iterative_deepcopy,
)
from .._impl import TrieInnerNode, TrieLeafNode


class AssignDestructionTimeYoungestPlusOneTriePostprocessor:
    """Functor to assign a destruction time property to trie nodes.

    Destruction time of leaf nodes are set to infinity. Destruction time of
    innner nodes is calculated as the minimum of its children's origin times
    plus one.
    """

    _assigned_property: str  # property name for the assigned destruction time.
    _origin_time_property: str  # property name for the node's origin time.

    def __init__(
        self: "AssignDestructionTimeYoungestPlusOneTriePostprocessor",
        assigned_property: str = "destruction_time",
        origin_time_property: str = "origin_time",
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        assigned_property : str, default "destruction_time"
            The property name for the assigned destruction tim.
        origin_time_property : str, default "origin_time"
            The property name for the node's origin time.
        """
        self._assigned_property = assigned_property
        self._origin_time_property = origin_time_property

    def __call__(
        self: "AssignDestructionTimeYoungestPlusOneTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Assign destruction times to trie nodes based on their origin times.

        Parameters
        ----------
        trie : TrieInnerNode
            The input trie to be postprocessed.
        p_differentia_collision : float
            The multiplicative inverse of the number of possible
            differentia.

            Not used in the current implementation.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie with assigned destruction times.
        """
        if not mutate:
            trie = anytree_iterative_deepcopy(
                trie, progress_wrap=progress_wrap
            )

        for node in progress_wrap(AnyTreeFastPreOrderIter(trie)):
            if node.is_leaf:
                setattr(node, self._assigned_property, math.inf)
                assert isinstance(node, TrieLeafNode)
            else:
                destruction_time = (
                    min(
                        getattr(child, self._origin_time_property)
                        for child in node.children
                    )
                    + 1
                )
                setattr(node, self._assigned_property, destruction_time)

            assert hasattr(node, self._assigned_property)

        return trie



"""Data structures to annotate genomes with."""

from . import stratum_ordered_stores
from ._HereditaryStratigraphicColumn import HereditaryStratigraphicColumn
from ._HereditaryStratigraphicColumnBundle import (
    HereditaryStratigraphicColumnBundle,
)
from ._HereditaryStratum import HereditaryStratum
from .stratum_ordered_stores import *  # noqa: F401

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "HereditaryStratigraphicColumn",
    "HereditaryStratigraphicColumnBundle",
    "HereditaryStratum",
] + stratum_ordered_stores.__all__

from .._auxiliary_lib import launder_impl_modules as _launder

_launder(
    [
        HereditaryStratigraphicColumn,
        HereditaryStratigraphicColumnBundle,
        HereditaryStratum,
    ],
    __name__,
)
del _launder  # prevent name from leaking



import numpy as np

from ._min_array_dtype import min_array_dtype


def as_compact_type(array: np.array) -> np.array:
    """
    Given an input numpy array, this function returns a copy of the input array
    cast to the most compact possible numpy datatype that can hold the values in the input array without overflow.

    Parameters
    ----------
    array : numpy.ndarray
        The input numpy array.

    Returns
    -------
    numpy.ndarray
        A copy of the input array cast to most compact possible numpy datatype
        that can hold the values in the input array without overflow.

    Notes
    -----
    Like `np.min_scalar_type` this method ignores loss of precision at values
    close to 0 (i.e., subnormal numbers).

    Examples
    --------
    >>> import numpy as np
    >>> from ._min_array_dtype import as_compact_type
    >>> arr = np.array([-1, 2, 3, 4, 5])
    >>> as_compact_type(arr)
    array([-1, 2, 3, 4, 5], dtype=int8)
    >>> arr = np.array([255, 256, 257])
    >>> as_compact_type(arr)
    array([255,   0,   1], dtype=uint8)
    """
    return array.astype(min_array_dtype(array))



import alifedata_phyloinformatics_convert as apc
import dendropy as dp


def tree_unweighted_robinson_foulds_distance(x, y) -> float:
    tree_a = apc.RosettaTree(x).as_dendropy
    tree_b = apc.RosettaTree(y).as_dendropy

    common_namespace = dp.TaxonNamespace()
    tree_a.migrate_taxon_namespace(common_namespace)
    tree_b.migrate_taxon_namespace(common_namespace)

    tree_a.encode_bipartitions()
    for bp in tree_a.bipartition_encoding:
        bp.is_mutable = False
    tree_b.encode_bipartitions()
    for bp in tree_b.bipartition_encoding:
        bp.is_mutable = False

    return dp.calculate.treecompare.unweighted_robinson_foulds_distance(
        tree_a, tree_b, is_bipartitions_updated=True
    )



import random

import pytest

from hstrat.hstrat import HereditaryStratigraphicColumn, stochastic_algo


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_policy_consistency(replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    spec = policy.GetSpec()
    instance = stochastic_algo.CalcNumStrataRetainedUpperBound(spec)
    column = HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
    )
    for num_strata_deposited in range(1, 10**3):
        policy_requirement = column.GetNumStrataRetained()
        for which in (
            instance,
            stochastic_algo.CalcNumStrataRetainedUpperBound(spec),
        ):
            assert (
                which(
                    policy,
                    num_strata_deposited,
                )
                >= policy_requirement
            )


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_eq(replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    spec = policy.GetSpec()
    instance = stochastic_algo.CalcNumStrataRetainedUpperBound(spec)

    assert instance == instance
    assert instance == stochastic_algo.CalcNumStrataRetainedUpperBound(
        spec,
    )
    assert instance is not None



import pandas as pd


def alifestd_chronological_sort(
    phylogeny_df: pd.DataFrame,
    how: str = "origin_time",
    mutate: bool = False,
) -> pd.DataFrame:
    """Sort rows so all organisms appear in chronological order, default
    `origin_time`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df.sort_values(by=how, inplace=True)
    phylogeny_df.reset_index(drop=True, inplace=True)
    return phylogeny_df



import numpy as np

from ._jit import jit


@jit(nopython=True)
def is_subset(subset: np.array, superset: np.array) -> bool:
    """Are all values in `subset` contained in `superset`?"""
    superset_lookup = set(superset)
    for val in subset:
        if val not in superset_lookup:
            return False

    return True



import typing

from ...._auxiliary_lib import (
    AnyTreeFastPreOrderIter,
    anytree_iterative_deepcopy,
)
from .._impl import TrieInnerNode


class AssignOriginTimeNodeRankTriePostprocessor:
    """Functor to assign trie nodes' rank as their the origin time."""

    _assigned_property: str

    def __init__(
        self: "AssignOriginTimeNodeRankTriePostprocessor",
        assigned_property: str = "origin_time",
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        assigned_property : str, default "origin_time"
            The property name for the assigned origin time.
        """
        self._assigned_property = assigned_property

    def __call__(
        self: "AssignOriginTimeNodeRankTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Assign origin times to trie nodes.

        Parameters
        ----------
        trie : TrieInnerNode
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.
            Not used in the current implementation.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie with assigned origin times.
        """
        if not mutate:
            trie = anytree_iterative_deepcopy(
                trie, progress_wrap=progress_wrap
            )

        for node in progress_wrap(AnyTreeFastPreOrderIter(trie)):
            setattr(node, self._assigned_property, node.rank)

        return trie



import typing

from ..frozen_instrumentation import HereditaryStratigraphicAssemblage
from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_to_specimen import col_to_specimen


def pop_to_assemblage(
    columns: typing.Iterable[HereditaryStratigraphicColumn],
    progress_wrap: typing.Callable = lambda x: x,
) -> HereditaryStratigraphicAssemblage:
    """Create a postprocessing representation of the differentia retained
    by a collection of HereditaryStratigraphicColumns.

    Parameters
    ----------
    columns : iterable of HereditaryStratigraphicColumn
        Data to serialize.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    """
    return HereditaryStratigraphicAssemblage(
        map(col_to_specimen, progress_wrap(columns))
    )



import typing


def policy_from_record(policy_record: str) -> typing.Any:
    """Parse a policy record string and return the corresponding policy object."""
    from ...stratum_retention_strategy import (  # noqa
        stratum_retention_algorithms as hstrat,
    )

    _ = hstrat  # mark as used to prevent autoflake purge
    return eval(policy_record)  # noqa



"""Functions that compare two columns.

Provides the foundation for phylogenetic inference tools.
"""

from ._calc_definitive_max_rank_of_first_retained_disparity_between import (
    calc_definitive_max_rank_of_first_retained_disparity_between,
)
from ._calc_definitive_max_rank_of_last_retained_commonality_between import (
    calc_definitive_max_rank_of_last_retained_commonality_between,
)
from ._calc_definitive_min_ranks_since_first_retained_disparity_with import (
    calc_definitive_min_ranks_since_first_retained_disparity_with,
)
from ._calc_definitive_min_ranks_since_last_retained_commonality_with import (
    calc_definitive_min_ranks_since_last_retained_commonality_with,
)
from ._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_probability_differentia_collision_between import (
    calc_probability_differentia_collision_between,
)
from ._calc_rank_of_first_retained_disparity_between import (
    calc_rank_of_first_retained_disparity_between,
)
from ._calc_rank_of_last_retained_commonality_between import (
    calc_rank_of_last_retained_commonality_between,
)
from ._calc_ranks_since_first_retained_disparity_with import (
    calc_ranks_since_first_retained_disparity_with,
)
from ._calc_ranks_since_last_retained_commonality_with import (
    calc_ranks_since_last_retained_commonality_with,
)
from ._diff_retained_ranks import diff_retained_ranks
from ._get_last_common_stratum_between import get_last_common_stratum_between
from ._get_nth_common_rank_between import get_nth_common_rank_between

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "calc_definitive_max_rank_of_first_retained_disparity_between",
    "calc_definitive_max_rank_of_last_retained_commonality_between",
    "calc_definitive_min_ranks_since_first_retained_disparity_with",
    "calc_definitive_min_ranks_since_last_retained_commonality_with",
    "calc_min_implausible_spurious_consecutive_differentia_collisions_between",
    "calc_probability_differentia_collision_between",
    "calc_rank_of_first_retained_disparity_between",
    "calc_rank_of_last_retained_commonality_between",
    "calc_ranks_since_first_retained_disparity_with",
    "calc_ranks_since_last_retained_commonality_with",
    "diff_retained_ranks",
    "get_last_common_stratum_between",
    "get_nth_common_rank_between",
]

from .._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



"""Implementation helpers."""

from ._AssignDestructionTimeYoungestPlusOneTriePostprocessor import (
    AssignDestructionTimeYoungestPlusOneTriePostprocessor,
)
from ._AssignOriginTimeExpectedValueTriePostprocessor import (
    AssignOriginTimeExpectedValueTriePostprocessor,
)
from ._AssignOriginTimeNaiveTriePostprocessor import (
    AssignOriginTimeNaiveTriePostprocessor,
)
from ._AssignOriginTimeNodeRankTriePostprocessor import (
    AssignOriginTimeNodeRankTriePostprocessor,
)
from ._CompoundTriePostprocessor import CompoundTriePostprocessor
from ._PeelBackConjoinedLeavesTriePostprocessor import (
    PeelBackConjoinedLeavesTriePostprocessor,
)
from ._SampleAncestralRollbacksTriePostprocessor import (
    SampleAncestralRollbacksTriePostprocessor,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "AssignDestructionTimeYoungestPlusOneTriePostprocessor",
    "AssignOriginTimeExpectedValueTriePostprocessor",
    "AssignOriginTimeNaiveTriePostprocessor",
    "AssignOriginTimeNodeRankTriePostprocessor",
    "CompoundTriePostprocessor",
    "PeelBackConjoinedLeavesTriePostprocessor",
    "SampleAncestralRollbacksTriePostprocessor",
]



import pandas as pd

from ._alifestd_find_chronological_inconsistency import (
    alifestd_find_chronological_inconsistency,
)


def alifestd_is_chronologically_ordered(phylogeny_df: pd.DataFrame) -> bool:
    """Are all organisms listed after members of their `ancestor_list`?

    Input dataframe is not mutated by this operation.
    """

    return alifestd_find_chronological_inconsistency(phylogeny_df) is None



"""Maintain constant size complexity with recency-proportional strata spacing.

The approximate space-filling MRCA-recency-proportional resolution policy
imposes an O(1) limit on the number of retained strata and guarantees that
retained strata will be exponentially distributed with respect to ranks elapsed
since their deposit. MRCA rank estimate uncertainty scales in the worst case
scales as O(n) with respect to the greater number of strata deposited on either
column. However, with respect to estimating the rank of the MRCA when lineages
diverged any fixed number of generations ago, uncertainty scales as O(n^{1/k}).

Under the MRCA-recency-proportional resolution policy, the number of strata
retained (i.e., space complexity) scales as O(1) with respect to the number of
strata deposited.

Suppose k is specified as the policy's target precision k. Then, the first k
strata deposited will be retained. Subsequently, strata are retained so that
MRCA rank estimate uncertainty is less than or equal to s * (1 - n^(-1/k)) is
the number of strata deposited and s is the true number of ranks deposited
since the MRCA. As n goes to infinity, the number of strata retained fluctuates
below a hard upper limit of 4k + 2 (inclusive) strata. For larger target space
utilizations, number of strata retained appears generally less than twice the
target space utilization.

See Also
--------
geom_seq_nth_root_tapered_algo:
    For a predicate retention policy that achieves the same guarantees for
    resolution but retains a constant column size exactly equal to the hard
    upper limit on number of strata retained.
"""

from ._Policy import Policy
from ._PolicySpec import PolicySpec
from ._enact._GenDropRanks import GenDropRanks
from ._enact._GenDropRanks_ import impls as _GenDropRanks_impls
from ._invar._CalcMrcaUncertaintyAbsUpperBound import (
    CalcMrcaUncertaintyAbsUpperBound,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBound import (
    CalcMrcaUncertaintyRelUpperBound,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRank,
)
from ._invar._CalcNumStrataRetainedUpperBound import (
    CalcNumStrataRetainedUpperBound,
)
from ._scry._CalcMrcaUncertaintyAbsExact import CalcMrcaUncertaintyAbsExact
from ._scry._CalcMrcaUncertaintyRelExact import CalcMrcaUncertaintyRelExact
from ._scry._CalcNumStrataRetainedExact import CalcNumStrataRetainedExact
from ._scry._CalcRankAtColumnIndex import CalcRankAtColumnIndex
from ._scry._IterRetainedRanks import IterRetainedRanks

__all__ = [
    "CalcMrcaUncertaintyAbsExact",
    "CalcMrcaUncertaintyAbsUpperBound",
    "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
    "CalcMrcaUncertaintyRelExact",
    "CalcMrcaUncertaintyRelUpperBound",
    "CalcMrcaUncertaintyRelUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
    "CalcNumStrataRetainedExact",
    "CalcNumStrataRetainedUpperBound",
    "CalcRankAtColumnIndex",
    "GenDropRanks",
    "_GenDropRanks_impls",
    "IterRetainedRanks",
    "Policy",
    "PolicySpec",
]

from ...._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import itertools as it

import numpy as np
import pytest

from hstrat.hstrat import depth_proportional_resolution_tapered_algo


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10**2,
        ),
        (2**32,),
    ],
)
def test_policy_consistency(depth_proportional_resolution, time_sequence):
    policy = depth_proportional_resolution_tapered_algo.Policy(
        depth_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = (
        depth_proportional_resolution_tapered_algo.CalcRankAtColumnIndex(spec)
    )
    for num_strata_deposited in time_sequence:
        for which in (
            instance,
            depth_proportional_resolution_tapered_algo.CalcRankAtColumnIndex(
                spec
            ),
        ):
            assert all(
                calculated == policy_requirement
                for policy_requirement, calculated in it.zip_longest(
                    policy.IterRetainedRanks(num_strata_deposited),
                    (
                        which(policy, i, num_strata_deposited)
                        for i in range(
                            policy.CalcNumStrataRetainedExact(
                                num_strata_deposited,
                            )
                        )
                    ),
                )
            )


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(depth_proportional_resolution):
    policy = depth_proportional_resolution_tapered_algo.Policy(
        depth_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = (
        depth_proportional_resolution_tapered_algo.CalcRankAtColumnIndex(spec)
    )

    assert instance == instance
    assert (
        instance
        == depth_proportional_resolution_tapered_algo.CalcRankAtColumnIndex(
            spec
        )
    )
    assert instance is not None



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_assign_root_ancestor_token,
    alifestd_find_root_ids,
    alifestd_to_working_format,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "root_ancestor_token",
    [
        "",
        "None",
        "none",
    ],
)
def test_alifestd_assign_root_ancestor_token(
    phylogeny_df, apply, root_ancestor_token
):
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df_ = phylogeny_df.copy()
    res_df = alifestd_assign_root_ancestor_token(
        phylogeny_df, root_ancestor_token
    )
    assert alifestd_validate(res_df)
    assert sorted(
        res_df[res_df["ancestor_list"] == f"[{root_ancestor_token}]"]["id"]
    ) == sorted(alifestd_find_root_ids(phylogeny_df_))
    assert phylogeny_df.equals(phylogeny_df_)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "root_ancestor_token",
    [
        "",
        "None",
        "none",
    ],
)
def test_alifestd_assign_root_ancestor_token_mutate(
    phylogeny_df, apply, root_ancestor_token
):
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df_ = phylogeny_df.copy()
    res_df = alifestd_assign_root_ancestor_token(
        phylogeny_df, root_ancestor_token, mutate=True
    )
    assert alifestd_validate(res_df)
    assert sorted(
        res_df[res_df["ancestor_list"] == f"[{root_ancestor_token}]"]["id"]
    ) == sorted(alifestd_find_root_ids(phylogeny_df_))



import typing

import numpy as np
import pandas as pd

from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._jit import jit
from ._jit_numba_dict_t import jit_numba_dict_t
from ._jit_numpy_int64_t import jit_numpy_int64_t


@jit(nopython=True)
def _reassign_ids(
    ids: np.array,
    ancestor_ids: np.array,
) -> typing.Tuple[typing.Dict[int, int], np.array]:
    reassignment = jit_numba_dict_t.empty(
        key_type=jit_numpy_int64_t,
        value_type=jit_numpy_int64_t,
    )
    for id_ in ids:
        reassignment[id_] = len(reassignment)

    return reassignment, np.array(
        [reassignment[ancestor_id] for ancestor_id in ancestor_ids]
    )


def alifestd_assign_contiguous_ids(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Reassign so each organism's id corresponds to its row number.

    Organisms retain the same row location; only id numbers change. Input
    dataframe is not mutated by this operation unless `mutate` True.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    reassignment, ancestor_ids = _reassign_ids(
        phylogeny_df["id"].to_numpy(dtype=np.int64),
        phylogeny_df["ancestor_id"].to_numpy(dtype=np.int64)
        if "ancestor_id" in phylogeny_df
        else np.array([], dtype=int),
    )

    phylogeny_df["id"] = np.arange(len(phylogeny_df))

    if len(ancestor_ids):
        phylogeny_df["ancestor_id"] = ancestor_ids
        phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
            phylogeny_df["id"], phylogeny_df["ancestor_id"]
        )
    else:
        phylogeny_df["ancestor_list"] = phylogeny_df["ancestor_list"].map(
            lambda ancestor_list_str: str(
                [
                    reassignment[ancestor_id]
                    for ancestor_id in alifestd_parse_ancestor_ids(
                        ancestor_list_str
                    )
                ]
            )
        )
        phylogeny_df.loc[
            phylogeny_df["ancestor_list"] == "[]", "ancestor_list"
        ] = "[none]"

    return phylogeny_df



import numpy as np

from hstrat._auxiliary_lib import jit, reversed_enumerate


def test_reversed_enumerate_list():
    l = [1, 2, 3, 4]
    assert list(reversed_enumerate(l)) == [(3, 4), (2, 3), (1, 2), (0, 1)]


def test_reversed_enumerate_str():
    s = "hello"
    assert list(reversed_enumerate(s)) == [
        (4, "o"),
        (3, "l"),
        (2, "l"),
        (1, "e"),
        (0, "h"),
    ]


def test_reversed_enumerate_empty():
    e = []
    assert list(reversed_enumerate(e)) == []


def test_reversed_enumerate_single_element():
    e = [1]
    assert list(reversed_enumerate(e)) == [(0, 1)]


def test_reversed_enumerate_tuple():
    t = (1, 2, 3)
    assert list(reversed_enumerate(t)) == [(2, 3), (1, 2), (0, 1)]


def test_reversed_enumerate_is_reverse():
    l = [1, 2, 3, 4]
    assert [*reversed_enumerate(l)] == [*reversed([*enumerate(l)])]


def test_reversed_enumerate_is_jitable():
    reversed_enumerate_jit = jit(nopython=True)(reversed_enumerate)
    a = np.array([1, 2, 3, 4])
    assert list(reversed_enumerate_jit(a)) == [(3, 4), (2, 3), (1, 2), (0, 1)]



import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_min_implausible_spurious_consecutive_differentia_collisions_between(
    wrap,
):
    assert (
        hstrat.calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=1
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=1
                )
            ),
            significance_level=0.51,
        )
        == 1
    )
    assert (
        hstrat.calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=1
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=1
                )
            ),
            significance_level=0.05,
        )
        == 5
    )
    assert (
        hstrat.calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=2
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=2
                )
            ),
            significance_level=0.05,
        )
        == 3
    )
    assert (
        hstrat.calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=5
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=5
                )
            ),
            significance_level=0.05,
        )
        == 1
    )
    assert (
        hstrat.calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=64
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=64
                )
            ),
            significance_level=0.51,
        )
        == 1
    )
    assert (
        hstrat.calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=64
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=64
                )
            ),
            significance_level=0.05,
        )
        == 1
    )
    assert (
        hstrat.calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=64
                )
            ),
            wrap(
                hstrat.HereditaryStratigraphicColumn(
                    stratum_differentia_bit_width=64
                )
            ),
            significance_level=0.01,
        )
        == 1
    )



import unittest

from hstrat._auxiliary_lib import zip_strict


class TestZipStrict(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_none(self):
        assert [*zip_strict()] == [*zip()]

    def test_single(self):
        assert [*zip_strict("")] == [*zip("")]
        assert [*zip_strict("a")] == [*zip("a")]
        assert [*zip_strict("ab")] == [*zip("ab")]

    def test_double(self):
        assert [*zip_strict("", range(0))] == [*zip("", range(0))]
        assert [*zip_strict("a", range(1))] == [*zip("a", range(1))]
        assert [*zip_strict("ab", range(2))] == [*zip("ab", range(2))]

        with self.assertRaises(ValueError):
            for a, b in zip(zip_strict("", range(1)), zip("", range(1))):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(zip_strict("a", range(0)), zip("a", range(0))):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(zip_strict("ab", range(1)), zip("ab", range(1))):
                assert a == b

    def test_triple(self):
        assert [*zip_strict("", range(0), [])] == [*zip("", range(0), [])]
        assert [*zip_strict("a", range(1), [None])] == [
            *zip("a", range(1), [None])
        ]
        assert [*zip_strict("ab", range(2), [None, None])] == [
            *zip("ab", range(2), [None, None])
        ]

        with self.assertRaises(ValueError):
            for a, b in zip(
                zip_strict("", range(1), []), zip("", range(1), [])
            ):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(
                zip_strict("", range(1), [None]), zip("", range(1), [None])
            ):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(
                zip_strict("a", range(0), [None]), zip("a", range(0), [None])
            ):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(
                zip_strict("a", range(0), []), zip("a", range(0), [])
            ):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(
                zip_strict("", range(0), [None]), zip("", range(0), [None])
            ):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(
                zip_strict("", range(1), [None]), zip("", range(1), [None])
            ):
                assert a == b

        with self.assertRaises(ValueError):
            for a, b in zip(
                zip_strict("ab", range(1), [None, None, None]),
                zip("ab", range(1), [None, None, None]),
            ):
                assert a == b


if __name__ == "__main__":
    unittest.main()



from copy import copy
import operator
import typing

from interval_search import binary_search

from .._HereditaryStratum import HereditaryStratum
from ._detail import HereditaryStratumOrderedStoreBase


class HereditaryStratumOrderedStoreList(HereditaryStratumOrderedStoreBase):
    """Interchangeable backing container for HereditaryStratigraphicColumn.

    Stores deposited strata using a list implementation. Retained strata are
    stored from most ancient (index 0, front) to most recent (back). Cloned
    stores instantiate an independent list (although strata are not deepcopied
    themselves).

    Potentially useful in scenarios where moderate strata counts are retained,
    many strata are deposited without column cloning, deleted strata tend to
    be more recent (i.e., not more ancient and toward the front of the list),
    or many comparisons to estimate most recent common ancestor are made
    between stratigraphic columns.
    """

    __slots__ = ("_data",)

    # strata stored from most ancient (index 0, front) to most recent (back)
    _data: typing.List[HereditaryStratum]

    def __init__(self: "HereditaryStratumOrderedStoreList"):
        """Initialize instance variables."""
        self._data = []

    def __eq__(
        self: "HereditaryStratumOrderedStoreList",
        other: "HereditaryStratumOrderedStoreList",
    ) -> bool:
        """Compare for value-wise equality."""
        # adapted from https://stackoverflow.com/a/4522896
        return (
            isinstance(
                other,
                self.__class__,
            )
            and self.__slots__ == other.__slots__
            and all(
                getter(self) == getter(other)
                for getter in [
                    operator.attrgetter(attr) for attr in self.__slots__
                ]
            )
        )

    def DepositStratum(
        self: "HereditaryStratumOrderedStoreList",
        rank: typing.Optional[int],
        stratum: "HereditaryStratum",
    ) -> None:
        """Insert a new stratum into the store.

        Parameters
        ----------
        rank : typing.Optional[int]
            The position of the stratum being deposited within the sequence of strata deposited into the column. Precisely, the number of strata that have been deposited before stratum.
        stratum : HereditaryStratum
            The stratum to deposit.
        """
        self._data.append(stratum)

    def GetNumStrataRetained(self: "HereditaryStratumOrderedStoreList") -> int:
        """How many strata are present in the store?

        May be fewer than the number of strata deposited if deletions have
        occured.
        """
        return len(self._data)

    def GetStratumAtColumnIndex(
        self: "HereditaryStratumOrderedStoreList",
        index: int,
        # needed for other implementations
        get_rank_at_column_index: typing.Optional[typing.Callable] = None,
    ) -> HereditaryStratum:
        """Get the stratum positioned at index i among retained strata.

        Index order is from most ancient (index 0) to most recent.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the index position within retained strata of
            the stratum deposited at rank r.
        """
        return self._data[index]

    def GetRankAtColumnIndex(
        self: "HereditaryStratumOrderedStoreList",
        index: int,
    ) -> int:
        """Map from deposition generation to column position.

        What is the deposition rank of the stratum positioned at index i
        among retained strata? Index order is from most ancient (index 0) to
        most recent.
        """
        res_rank = self.GetStratumAtColumnIndex(index).GetDepositionRank()
        assert res_rank is not None
        return res_rank

    def GetColumnIndexOfRank(
        self: "HereditaryStratumOrderedStoreList",
        rank: int,
    ) -> typing.Optional[int]:
        """Map from column position to deposition generation

        What is the index position within retained strata of the stratum
        deposited at rank r? Returns None if no stratum with rank r is present
        within the store.
        """
        if self.GetNumStrataRetained() == 0:
            return None
        else:
            res_idx = binary_search(
                lambda idx: self.GetRankAtColumnIndex(idx) >= rank,
                0,
                self.GetNumStrataRetained() - 1,
            )
            if res_idx is None:
                return None
            elif self.GetRankAtColumnIndex(res_idx) == rank:
                return res_idx
            else:
                return None

    def DelRanks(
        self: "HereditaryStratumOrderedStoreList",
        ranks: typing.Iterator[int],
        # deposition ranks might not be stored in strata
        get_column_index_of_rank: typing.Optional[typing.Callable] = None,
    ) -> None:
        """Purge strata with specified deposition ranks from the store.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the deposition rank of the stratum positioned
            at index i among retained strata.
        """
        if get_column_index_of_rank is None:
            get_column_index_of_rank = self.GetColumnIndexOfRank

        indices = [get_column_index_of_rank(rank) for rank in ranks]
        # adapted from https://stackoverflow.com/a/11303234/17332200
        # iterate over indices in reverse order to prevent invalidation
        # reversed() is an potential optimization
        # given indices is assumed to be in ascending order
        for index in sorted(reversed(indices), reverse=True):
            assert index is not None
            del self._data[index]

    def IterRetainedRanks(
        self: "HereditaryStratumOrderedStoreList",
    ) -> typing.Iterator[int]:
        """Iterate over deposition ranks of strata present in the store from
        most ancient to most recent.

        The store may be altered during iteration without iterator
        invalidation, although subsequent updates will not be reflected in the
        iterator.
        """
        # must make copy to prevent invalidation when strata are deleted
        # note, however, that copy is made lazily
        # (only when first item requested)
        ranks = [stratum.GetDepositionRank() for stratum in self._data]
        for rank in ranks:
            assert rank is not None
            yield rank

    def IterRetainedStrata(
        self: "HereditaryStratumOrderedStoreList",
    ) -> typing.Iterator[HereditaryStratum]:
        """Iterate over stored strata from most ancient to most recent."""
        yield from self._data

    def IterRankDifferentiaZip(
        self: "HereditaryStratumOrderedStoreList",
        # deposition ranks might not be stored in strata
        get_rank_at_column_index: typing.Optional[typing.Callable] = None,
        start_column_index: int = 0,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over differentia and corresponding deposition ranks.

        Values yielded as tuples. Guaranteed ordered from most ancient to most
        recent.

        Parameters
        ----------
        get_rank_at_column_index : callable, optional
            Callable that returns the deposition rank of the stratum positioned
            at index i among retained strata.
        start_column_index : callable, optional
            Number of strata to skip over before yielding first result from the
            iterator. Default 0, meaning no strata are skipped over.
        """
        if get_rank_at_column_index is None:
            get_rank_at_column_index = self.GetRankAtColumnIndex

        # adapted from https://stackoverflow.com/a/12911454
        for index in range(start_column_index, len(self._data)):
            stratum = self._data[index]
            yield (get_rank_at_column_index(index), stratum.GetDifferentia())

    def Clone(
        self: "HereditaryStratumOrderedStoreList",
    ) -> "HereditaryStratumOrderedStoreList":
        """Create an independent copy of the store.

        Returned copy contains identical data but may be freely altered without
        affecting data within this store.
        """
        # shallow copy
        result = copy(self)
        # do semi-shallow clone on select elements
        # see https://stackoverflow.com/a/47859483 for performance consierations
        result._data = [*self._data]
        return result



"""Retains strata probabilistically.

It would be a poor choice to use in practice because mismatches between the
particular ranks that each column happens to have strata for will degrade
the effectiveness of comparisons between columns. Rather, it is included in
the library as an edge case for testing purposes. Worst-case MRCA rank estimate uncertainty scales as O(n) with respect to the greater number of strata deposited on either column being compared.

Under the pseudostochastic resolution policy, the worst and average case number
of strata retained (i.e., space complexity) scales as O(n) with respect to
the number of strata deposited.

This policy implementation that the most ancient and most recent strata will
always be retained. For the secondmost recently deposited sratum, a
pseudorandom coin flip is performed. Depending on the outcome of that coin
flip, the stratum is either immediately purged or retained permanently.
"""

from ._Policy import Policy
from ._PolicySpec import PolicySpec
from ._enact._GenDropRanks import GenDropRanks
from ._enact._GenDropRanks_ import impls as _GenDropRanks_impls
from ._invar._CalcMrcaUncertaintyAbsUpperBound import (
    CalcMrcaUncertaintyAbsUpperBound,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBound import (
    CalcMrcaUncertaintyRelUpperBound,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRank,
)
from ._invar._CalcNumStrataRetainedUpperBound import (
    CalcNumStrataRetainedUpperBound,
)
from ._scry._CalcMrcaUncertaintyAbsExact import CalcMrcaUncertaintyAbsExact
from ._scry._CalcMrcaUncertaintyRelExact import CalcMrcaUncertaintyRelExact
from ._scry._CalcNumStrataRetainedExact import CalcNumStrataRetainedExact
from ._scry._CalcRankAtColumnIndex import CalcRankAtColumnIndex
from ._scry._IterRetainedRanks import IterRetainedRanks

__all__ = [
    "CalcMrcaUncertaintyAbsExact",
    "CalcMrcaUncertaintyAbsUpperBound",
    "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
    "CalcMrcaUncertaintyRelExact",
    "CalcMrcaUncertaintyRelUpperBound",
    "CalcMrcaUncertaintyRelUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
    "CalcNumStrataRetainedExact",
    "CalcNumStrataRetainedUpperBound",
    "CalcRankAtColumnIndex",
    "GenDropRanks",
    "_GenDropRanks_impls",
    "IterRetainedRanks",
    "Policy",
    "PolicySpec",
]

from ...._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import pytest

from hstrat._auxiliary_lib import intersect_ranges


@pytest.mark.parametrize(
    "ranges, expected",
    [
        ([(0, 50), (3, 9), (7, 10)], (7, 9)),
        ([(1, 3), (3, 7), (3, 9)], None),
        ([(1, 5), (6, 8), (9, 12)], None),
        ([], None),
        ([(1, 2)], (1, 2)),
        ([(1, 1)], None),
        ([(1, 1), (2, 2), (3, 3)], None),
    ],
)
def test_intersect_ranges(ranges, expected):
    assert intersect_ranges(ranges) == expected



"""Top-level package for hstrat."""

__author__ = """Matthew Andres Moreno"""
__email__ = "m.more500@gmail.com"
__version__ = "1.8.5"

from . import (
    _auxiliary_lib,
    frozen_instrumentation,
    genome_instrumentation,
    juxtaposition,
    phylogenetic_inference,
    serialization,
    stratum_retention_strategy,
    stratum_retention_viz,
    test_drive,
)

__all__ = [
    "_auxiliary_lib",
    "frozen_instrumentation",
    "genome_instrumentation",
    "juxtaposition",
    "phylogenetic_inference",
    "serialization",
    "stratum_retention_strategy",
    "stratum_retention_viz",
    "test_drive",
]



import pandas as pd

from hstrat._auxiliary_lib import (
    alifestd_has_increasing_ids,
    alifestd_make_empty,
)


def test_empty():
    assert alifestd_has_increasing_ids(alifestd_make_empty())


def test_singleton():
    df = pd.DataFrame({"id": [2], "ancestor_id": [2]})
    assert alifestd_has_increasing_ids(df)

    df = pd.DataFrame({"id": [0], "ancestor_list": ["[None]"]})
    assert alifestd_has_increasing_ids(df)


def test_ancestor_id_column_present():
    df = pd.DataFrame({"id": [2, 3, 4, 5], "ancestor_id": [2, 2, 3, 4]})
    assert alifestd_has_increasing_ids(df)

    df_false_case = pd.DataFrame(
        {"id": [2, 3, 4, 5], "ancestor_id": [2, 2, 5, 2]}
    )
    assert not alifestd_has_increasing_ids(df_false_case)


def test_ancestor_list_column_present():
    df = pd.DataFrame(
        {"id": [2, 3, 4, 5], "ancestor_list": ["[2]", "[2]", "[3,2]", "[4,3]"]}
    )
    assert alifestd_has_increasing_ids(df)

    df_false_case = pd.DataFrame(
        {"id": [2, 3, 4, 5], "ancestor_list": ["[2]", "[2,4]", "[2]", "[4,3]"]}
    )
    assert not alifestd_has_increasing_ids(df_false_case)



from hstrat._auxiliary_lib import jit, reversed_range


def test_reversed_range_jit():
    reversed_range_jit = jit(nopython=True)(reversed_range)
    # Test with stop value specified
    result = list(reversed_range_jit(1, 5))
    assert result == [*reversed(range(1, 5))]
    result = list(reversed_range_jit(0, 10))
    assert result == [*reversed(range(0, 10))]

    # Test without stop value specified
    result = list(reversed_range_jit(5))
    assert result == [*reversed(range(5))]
    result = list(reversed_range_jit(0))
    assert result == [*reversed(range(0))]


def test_reversed_range_stop_specified():
    # Test with stop value specified
    result = list(reversed_range(1, 5))
    assert result == [*reversed(range(1, 5))]
    result = list(reversed_range(0, 10))
    assert result == [*reversed(range(0, 10))]


def test_reversed_range_no_stop_specified():
    # Test without stop value specified
    result = list(reversed_range(5))
    assert result == [*reversed(range(5))]
    result = list(reversed_range(0))
    assert result == [*reversed(range(0))]


def test_reversed_range_single_element():
    # Test with single-element range
    result = list(reversed_range(5, 6))
    assert result == [*reversed(range(5, 6))]
    result = list(reversed_range(5, 5))
    assert result == [*reversed(range(5, 5))]


def test_reversed_range_negative_values():
    # Test with negative start and stop values
    result = list(reversed_range(-5, -1))
    assert result == [*reversed(range(-5, -1))]


def test_reversed_range_empty_range():
    # Test with an empty range
    result = list(reversed_range(5, 5))
    assert result == [*reversed(range(5, 5))]
    result = list(reversed_range(5, 6))
    assert result == [*reversed(range(5, 6))]
    result = list(reversed_range(0, 0))
    assert result == [*reversed(range(0, 0))]



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact


def get_nth_common_rank_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    n: int,
) -> typing.Optional[int]:
    """Return the nth rank retained by both columns.

    Zero indexed. Returns None if n + 1 common ranks do not exist between
    first and second.
    """
    assert n >= 0

    # helper setup
    first_iter = first.IterRetainedRanks()
    second_iter = second.IterRetainedRanks()
    first_cur_rank = next(first_iter)
    second_cur_rank = next(second_iter)

    try:
        while True:
            if first_cur_rank == second_cur_rank:
                # strata at common rank
                if n == 0:
                    return first_cur_rank

                n -= 1
                # advance first
                first_cur_rank = next(first_iter)
                # advance second
                second_cur_rank = next(second_iter)
            elif first_cur_rank < second_cur_rank:
                # current stratum on first column older than on second column
                # advance to next-newer stratum on first column
                first_cur_rank = next(first_iter)
            elif first_cur_rank > second_cur_rank:
                # current stratum on second column older than on first column
                # advance to next-newer stratum on second column
                second_cur_rank = next(second_iter)

    except StopIteration:
        return None



import os

import pytest

from hstrat import hstrat

from . import _impl as impl

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
def test_postprocessing_consistency(orig_tree, retention_policy):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=1,
        ).CloneNthDescendant(num_depositions),
    )

    nop_df, rb_df1, rb_df2 = hstrat.build_tree_trie_ensemble(
        extant_population,
        trie_postprocessors=[
            lambda x, **kwargs: x,
            hstrat.SampleAncestralRollbacksTriePostprocessor(seed=1),
            hstrat.SampleAncestralRollbacksTriePostprocessor(seed=1),
        ],
    )

    assert not nop_df.equals(rb_df1)
    assert rb_df1.equals(rb_df2)
